###
# Copyright 2020 Hewlett Packard Enterprise, Inc. All rights reserved.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#  http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
###

# -*- coding: utf-8 -*-
"""Direct module for working with Redfish/REST technology."""

# ---------Imports---------

import base64
import hashlib
import logging
import sys
from six.moves.urllib.parse import urlparse

from redfish.rest.connections import (
    Blobstore2Connection,
    HttpConnection,
    InvalidCredentialsError,
    OneTimePasscodeError,
    UnauthorizedLoginAttemptError,
    TokenExpiredError,
)

# ---------End of imports---------


# ---------Debug logger---------

LOGGER = logging.getLogger(__name__)


# ---------End of debug logger---------


class ServerDownOrUnreachableError(Exception):
    """Raised when server is unreachable."""

    pass


class JsonDecodingError(Exception):
    """Raised when there is an error in json data."""

    pass


class AuthMethod(object):
    """AUTH Method class

    **BASIC**, **SESSION**, **CERTIFICATE** variables translate to their string counterparts
    `basic`, `session`, `certificate`."""

    BASIC = "basic"
    SESSION = "session"
    CERTIFICATE = "certificate"


class RestClientBase(object):
    """Base REST client. Each RestClientBase has a connection object built by parsing the
    client_kwargs. This connection is used for communicating remotely or locally.

    :param biospassword: The iLO Gen9 bios password. See :func:`bios_password`
    :type biospassword: str
    :param \\**client_kwargs: Arguments to pass to the client argument. For possible values see
                              :mod:`redfish.rest.connections.Blobstore2Connection` for a local
                              connection or :mod:`redfish.rest.connections.HttpConnection`
                              for remote connection."""

    def __init__(self, biospassword=None, **client_kwargs):
        self.connection = None
        self._biospassword = biospassword
        self._build_connection(**client_kwargs)

    @property
    def bios_password(self):
        """Property for the biospassword. Only required on Gen9 iLO 4 when RBSU bios password is set
        and modifying bios settings
        """
        return self._biospassword

    @bios_password.setter
    def bios_password(self, bios_pw):
        """set the bios password"""
        self._biospassword = bios_pw

    def _build_connection(self, **conn_kwargs):
        """Build the appropriate connection for the client"""
        base_url = conn_kwargs.pop("base_url", None)
        if not base_url or base_url.startswith("blobstore://"):
            self.connection = Blobstore2Connection(**conn_kwargs)
        elif not base_url.startswith("https://"):
            base_url = "https://" + base_url
            _ = conn_kwargs.pop("username", None)
            _ = conn_kwargs.pop("password", None)
            _ = conn_kwargs.pop("sessionid", None)
            _ = conn_kwargs.pop("login_otp", None)
            self.connection = HttpConnection(base_url, self._cert_data, **conn_kwargs)
        else:
            _ = conn_kwargs.pop("username", None)
            _ = conn_kwargs.pop("password", None)
            _ = conn_kwargs.pop("sessionid", None)
            _ = conn_kwargs.pop("login_otp", None)
            self.connection = HttpConnection(base_url, self._cert_data, **conn_kwargs)

    def _get_req_headers(self, headers=None):
        """Base _get_req_headers function"""
        return headers if headers else {}

    def get(self, path, args=None, headers=None):
        """Perform a GET request

        :param path: The URI path.
        :type path: str
        :param args: Any query to add to the URI. (Can also be directly added to the URI)
        :type args: dict
        :param headers: Any extra headers to add to the request.
        :type headers: dict
        :returns: A :class:`redfish.rest.containers.RestResponse` object
        """
        try:
            return self.connection.rest_request(
                path,
                method="GET",
                args=args,
                headers=self._get_req_headers(headers=headers),
            )
        except ValueError:
            LOGGER.debug("Error in json object getting path: %s", path)
            raise JsonDecodingError("Error in json decoding.")

    def patch(self, path, body, args=None, headers=None):
        """Perform a PATCH request

        :param path: The URI path.
        :type path: str
        :param body: The body to pass with the request.
        :type body: dict
        :param args: Any query to add to the URI. (Can also be directly added to the URI)
        :type args: dict
        :param headers: Any extra headers to add to the request.
        :type headers: dict
        :returns: A :class:`redfish.rest.containers.RestResponse` object
        """
        return self.connection.rest_request(
            path,
            body=body,
            method="PATCH",
            args=args,
            headers=self._get_req_headers(headers=headers),
        )

    def post(self, path, body, args=None, headers=None):
        """Perform a POST request

        :param path: The URI path.
        :type path: str
        :param body: The body to pass with the request.
        :type body: dict
        :param args: Any query to add to the URI. (Can also be directly added to the URI)
        :type args: dict
        :param headers: Any extra headers to add to the request.
        :type headers: dict
        :returns: A :class:`redfish.rest.containers.RestResponse` object
        """
        return self.connection.rest_request(
            path,
            body=body,
            method="POST",
            args=args,
            headers=self._get_req_headers(headers=headers),
        )

    def put(self, path, body, args=None, headers=None):
        """Perform a PUT request

        :param path: The URI path.
        :type path: str
        :param body: The body to pass with the request.
        :type body: dict
        :param args: Any query to add to the URI. (Can also be directly added to the URI)
        :type args: dict
        :param headers: Any extra headers to add to the request.
        :type headers: dict
        :returns: A :class:`redfish.rest.containers.RestResponse` object
        """
        return self.connection.rest_request(
            path,
            body=body,
            method="PUT",
            args=args,
            headers=self._get_req_headers(headers=headers),
        )

    def head(self, path, headers=None):
        """Perform a HEAD request

        :param path: The URI path.
        :type path: str
        :param headers: Any extra headers to add to the request.
        :type headers: dict
        :returns: A :class:`redfish.rest.containers.RestResponse` object
        """
        return self.connection.rest_request(path, method="HEAD", headers=self._get_req_headers(headers=headers))

    def delete(self, path, headers=None):
        """Perform a DELETE request

        :param path: The URI path.
        :type path: str
        :param args: Any query to add to the URI. (Can also be directly added to the URI)
        :type args: dict
        :returns: A :class:`redfish.rest.containers.RestResponse` object
        """
        return self.connection.rest_request(path, method="DELETE", headers=self._get_req_headers(headers=headers))


class RestClient(RestClientBase):
    """REST client with Redfish and LegacyRest support built on top.

    :param default_prefix: The root of the REST API, either /redfish/v1/ or /rest/v1.
    :type default_prefix: str
    :param is_redfish: Flag to force redfish conformance, even on LegacyRest clients.
    :type is_redfish: bool
    :param username: The username of the account to login with.
    :type username: str
    :param password: The password of the account to login with.
    :type password: str
    :param auth: The authentication type to force.
    :type auth: str or :class:`AuthMethod` class variable.
    :param ca_cert_data: Dictionary containing the certificate authority data (user CA, \
                         user root CA, user root CA key
    :type ca_cert_data: dict
    :param \\**client_kwargs: Arguments to create a :class:`RestClientBase` instance.
    """

    def __init__(
        self,
        default_prefix="/redfish/v1/",
        is_redfish=True,
        username=None,
        password=None,
        sessionid=None,
        base_url=None,
        auth=None,
        ca_cert_data=None,
        login_otp=None,
        **client_kwargs
    ):
        """Create a Rest Client object"""
        self.default_prefix = default_prefix
        self.is_redfish = is_redfish
        self.root = None
        self.auth_type = self._get_auth_type(auth, ca_cert_data=ca_cert_data, **client_kwargs)
        self._auth_key = None
        self._user_pass = (username, password)
        self._session_location = None
        self._cert_data = ca_cert_data
        self.login_otp = login_otp
        super(RestClient, self).__init__(
            username=username, password=password, sessionid=sessionid, base_url=base_url, login_otp=login_otp, **client_kwargs
        )

    def __enter__(self):
        """Create a connection and return the session object"""
        self.login()
        return self

    def __exit__(self, exc_type, exc_val, exc_tb):
        """Close the connection"""
        self.logout()

    def _get_auth_type(self, auth_param, ca_cert_data=None, **client_kwargs):
        """Get the auth type based on key args or positional argument.
        Defaults to session auth."""
        if not auth_param:
            # _ca_cert_data = client_kwargs.get('ca_cert_data')
            if ca_cert_data:
                if ("ca_certs" in ca_cert_data and ca_cert_data["ca_certs"]) or (
                    "cert_file" in ca_cert_data and ca_cert_data["cert_file"]
                ):
                    if (ca_cert_data.get("cert_file") and ca_cert_data.get("key_file")) or ca_cert_data.get("ca_certs"):
                        return AuthMethod.CERTIFICATE
            return AuthMethod.SESSION

        return auth_param

    @property
    def base_url(self):
        """The connection's URL to make calls against"""
        return self.connection.base_url

    @property
    def proxy(self):
        """The connection's proxy, if any."""
        try:
            proxy = self.connection.proxy
        except AttributeError:
            proxy = None
        return proxy

    @property
    def session_key(self):
        """The Client's session key, if any."""
        return self._auth_key if self.auth_type in [AuthMethod.SESSION, AuthMethod.CERTIFICATE] else None

    @session_key.setter
    def session_key(self, ses_key):
        """Set _auth_key to a session key"""
        self._auth_key = ses_key

    @property
    def basic_auth(self):
        """The Client's basic auth header, if any."""
        return self._auth_key if self.auth_type == AuthMethod.BASIC else None

    @basic_auth.setter
    def basic_auth(self, bas_key):
        """Set _auth_key to a basic auth header"""
        self._auth_key = bas_key

    @property
    def session_location(self):
        """The session URI. Used for deleting the session when we logout."""
        session_loc = None
        if self._session_location:
            if self.base_url == "blobstore://.":
                session_loc = self._session_location.replace("https://", "")
                session_loc = session_loc.replace(" ", "%20")
            else:
                parse_object = urlparse(self.base_url)
                if parse_object.hostname is not None:
                    newurl = "https://" + parse_object.hostname
                    session_loc = self._session_location.replace(newurl, "")
                else:
                    session_loc = self._session_location.replace(self.base_url, "")
        return session_loc

    @session_location.setter
    def session_location(self, ses_loc):
        """Set the session URI"""
        login_url_p = urlparse(self.base_url)
        new_port = login_url_p.port
        # The session_url returned by iLO does not include the port information
        # If we're using a non-standard port to connect to iLO we have to
        # manually insert it, based on the base_url.
        if new_port and ses_loc:
            session_location_p = urlparse(ses_loc)
            session_location_p = session_location_p._replace(
                netloc="{}:{}".format(session_location_p.hostname, new_port)
            )
            self._session_location = session_location_p.geturl()
        else:
            self._session_location = ses_loc

    @property
    def username(self):
        """The username, if any. Once a login function has been called the credentials are removed
        from memory for security and this will return None."""
        return self._user_pass[0]

    @property
    def password(self):
        """The password, if any. Once a login function has been called the credentials are removed
        from memory for security and this will return None."""
        return self._user_pass[1]

    @property
    def login_url(self):
        """The login URI from the root response. This is where we post the
        credentials for a login."""
        login_url = None

        try:
            login_url = self.root.obj.Links.Sessions["@odata.id"]
        except KeyError:
            login_url = self.root.obj.links.Sessions.href
        finally:
            if not login_url:
                raise ServerDownOrUnreachableError("Cannot locate the login url. Is this a Rest or" " Redfish server?")
        return login_url

    def login(self, auth=AuthMethod.SESSION):
        """Login to a Redfish or LegacyRest server.
        If auth is not supplied login will intelligently
        choose the authentication mode based on the arguments passed.
        Basic authentication MUST be specified with `auth`.

        :param auth: The auth type to login with.
        :type auth: str or :class:`AuthMethod` class variable
        """
        if auth:
            auth = self._get_auth_type(auth)
            self.auth_type = auth

        if self.auth_type is AuthMethod.BASIC:
            self._get_root()
            self._basic_login()
        elif self.auth_type is AuthMethod.SESSION:
            self._get_root()
            self._session_login()
        elif self.auth_type is AuthMethod.CERTIFICATE:
            self._cert_login()
            self._get_root()

    def logout(self):
        """Logout of session.

        YOU MUST CALL THIS WHEN YOU ARE DONE TO FREE UP SESSIONS"""
        if self.session_location:
            resp = self.delete(self.session_location)
            LOGGER.info("User logged out: %s", resp.read)
        self._auth_key = None
        self.session_location = None

    def _get_root(self):
        """Get the root response of the server"""
        if not self.root:
            resp = self.get(self.default_prefix)

            if resp.status != 200:
                raise ServerDownOrUnreachableError("Server not reachable, " "return code: %d" % resp.status)
            self.root = resp

    def _basic_login(self):
        """Login using basic authentication"""
        LOGGER.info("Performing basic authentication.")
        if not self.basic_auth:
            auth_key = base64.b64encode(("{}:{}".format(self.username, self.password)).encode("utf-8")).decode("utf-8")
            self.basic_auth = "Basic {}".format(auth_key)

        headers = dict()
        headers["Authorization"] = self.basic_auth

        respvalidate = self.get(self.login_url, headers=headers)

        if respvalidate.status == 401:
            self._credential_err()
        else:
            self._user_pass = (None, None)

    def _session_login(self):
        """Login using session authentication"""
        if not self.connection.session_key:
            LOGGER.info("Performing session authentication.")
            data = dict()
            data["UserName"] = self.username
            data["Password"] = self.password
            if self.login_otp is not None:
                data["Token"] = self.login_otp

            headers = dict()
            resp = self.post(self.login_url, body=data, headers=headers)
            try:
                respread = resp.read
                respread = respread.replace("\\\\", "\\")
                # LOGGER.info("%s" % respread)
            except ValueError:
                pass
            LOGGER.debug("Login returned code %s: %s", resp.status, respread)

            self.session_key = resp.session_key
            self.session_location = resp.session_location
            self.login_return_code = resp.status
            self.login_response = respread
        else:
            self.session_key = self.connection.session_key

        if hasattr(self, "login_response") and self.login_response:
            if "OneTimePasscodeSent" in self.login_response:
                raise OneTimePasscodeError()
            elif "UnauthorizedLogin" in self.login_response:
                raise UnauthorizedLoginAttemptError("Error " + str(
                    self.login_return_code) + ". Login is unauthorized.\nPlease check the credentials/OTP entered.\n")
            elif "TokenExpired" in self.login_response:
                raise TokenExpiredError("Error " + str(
                    self.login_return_code) + ". The OTP entered has expired. Please enter credentials again.\n")
        elif not self.session_key and not resp.status == 200:
            self._credential_err()
        else:
            self._user_pass = (None, None)

    def _cert_login(self):
        """Login using certificate authentication"""
        self.session_key, self.session_location = self.connection.cert_login()

    def _credential_err(self):
        """Return credential error based on delay"""
        try:
            if self.is_redfish:
                delay = self.root.obj.Oem.Hpe.Sessions.LoginFailureDelay
            else:
                delay = self.root.obj.Oem.Hp.Sessions.LoginFailureDelay
        except KeyError:
            delay = 5

        raise InvalidCredentialsError(delay)

    def _get_req_headers(self, headers=None, optionalpassword=None):
        """Get the request headers

        :param headers: additional headers to be utilized
        :type headers: str
        :param optionalpassword: provide password for authentication.
        :type optionalpassword: str.
        :returns: returns headers
        """
        headers = headers if isinstance(headers, dict) else dict()
        h_list = [header.lower() for header in headers]
        auth_headers = True if "x-auth-token" in h_list or "authorization" in h_list else False

        token = self._biospassword if self._biospassword else optionalpassword
        if token:
            token = optionalpassword.encode("utf-8") if type(optionalpassword).__name__ in "basestr" else token
            hash_object = hashlib.new("SHA256")
            hash_object.update(token)
            headers["X-HPRESTFULAPI-AuthToken"] = hash_object.hexdigest().upper()

        if self.session_key and not auth_headers:
            headers["X-Auth-Token"] = self.session_key
        elif self.basic_auth and not auth_headers:
            headers["Authorization"] = self.basic_auth

        if self.is_redfish:
            headers["OData-Version"] = "4.0"

        return headers

    def get_resource_directory(self):
        try:
            resource_uri = self.root.obj.Oem.Hpe.Links.ResourceDirectory["@odata.id"]
        except KeyError:
            sys.stderr.write("Resource directory is only available on HPE servers.\n")
            return None

        response = self.get(resource_uri)
        resources = []

        if response.status == 200:
            sys.stdout.write("\tFound resource directory at /redfish/v1/resourcedirectory" + "\n\n")
            resources = response.dict["Instances"]
        else:
            sys.stderr.write("\tResource directory missing at /redfish/v1/resourcedirectory" + "\n")

        return resources

    def get_gen(self):
        rootresp = self.root.obj
        # Default iLO 5
        ilogen = 5
        iloversion = None
        gencompany = next(iter(rootresp.get("Oem", {}).keys()), None) in ("Hpe", "Hp")
        comp = "Hp" if gencompany else None
        comp = "Hpe" if rootresp.get("Oem", {}).get("Hpe", None) else comp
        if comp and next(iter(rootresp.get("Oem", {}).get(comp, {}).get("Manager", {}))).get("ManagerType", None):
            ilogen = next(iter(rootresp.get("Oem", {}).get(comp, {}).get("Manager", {}))).get("ManagerType")
            ilover = next(iter(rootresp.get("Oem", {}).get(comp, {}).get("Manager", {}))).get("ManagerFirmwareVersion")
            if ilogen.split(" ")[-1] == "CM":
                # Assume iLO 4 types in Moonshot
                ilogen = 4
                iloversion = None
            else:
                ilogen = ilogen.split(" ")[1]
                iloversion = float(ilogen.split(" ")[-1] + "." + "".join(ilover.split(".")))
        return (ilogen, iloversion)


class LegacyRestClient(RestClient):
    """Class for a **Legacy REST** client instance.
    Instantiates appropriate Rest object based on existing configuration.
    Use this to retrieve a pre-configured Legacy Rest Class.

    Basic arguments include (These can be omitted for most local connections):

    * **base_url**: The IP or Hostname of the server to perform operations on.
    * **username**: The username of the account to login with.
    * **password**: The username of the account to login with.

    For full description of the arguments allowed see :class:`RestClient`"""

    def __init__(self, **client_kwargs):
        kwargs = {
            "default_prefix": client_kwargs.pop("default_prefix", "/rest/v1"),
            "is_redfish": client_kwargs.pop("is_redfish", False),
        }
        kwargs.update(client_kwargs)
        super(LegacyRestClient, self).__init__(**kwargs)


class RedfishClient(RestClient):
    """Class for a **Redfish** client instance.
    Instantiates appropriate Redfish object based on existing configuration.
    Use this to retrieve a pre-configured Redfish Class.

    Basic arguments include (These can be omitted for most local connections):

    * **base_url**: The IP or Hostname of the server to perform operations on. None for local.
    * **username**: The username of the account to login with.
    * **password**: The username of the account to login with.

    For full description of the arguments allowed see :class:`RestClient`"""

    def __init__(self, **client_kwargs):
        kwargs = {
            "default_prefix": client_kwargs.pop("default_prefix", "/redfish/v1"),
            "is_redfish": client_kwargs.pop("is_redfish", True),
        }
        kwargs.update(client_kwargs)
        super(RedfishClient, self).__init__(**kwargs)



 # Copyright 2020 Hewlett Packard Enterprise Development LP
 #
 # Licensed under the Apache License, Version 2.0 (the "License"); you may
 # not use this file except in compliance with the License. You may obtain
 # a copy of the License at
 #
 #      http://www.apache.org/licenses/LICENSE-2.0
 #
 # Unless required by applicable law or agreed to in writing, software
 # distributed under the License is distributed on an "AS IS" BASIS, WITHOUT
 # WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. See the
 # License for the specific language governing permissions and limitations
 # under the License.

# -*- coding: utf-8 -*-
"""
An example of uploading a firmware component with a size less than
32 GB, and associated compsig signature file to the iLO Repository.
"""

import os
import sys
import json
from redfish import RedfishClient
from redfish.rest.v1 import ServerDownOrUnreachableError

from get_resource_directory import get_resource_directory

def upload_firmware(_redfishobj, firmware_loc, compsig_loc, update_repo=True, update_target=False):
    resource_instances = get_resource_directory(_redfishobj)

    if DISABLE_RESOURCE_DIR or not resource_instances:
        #resource directory is not available so we will navigate through paths manually
        update_service_uri = _redfishobj.root.obj['UpdateService']['@odata.id']
    else:
        #obtain all account instances from resource directory
        for instance in resource_instances:
            if '#UpdateService.' in instance['@odata.type']:
                update_service_uri = instance['@odata.id']

    update_service_response = _redfishobj.get(update_service_uri)

    path = update_service_response.obj.HttpPushUri

    body = []
    json_data = {'UpdateRepository': update_repo, 'UpdateTarget': update_target, 'ETag': 'atag', 'Section': 0}
    session_key = _redfishobj.session_key

    filename = os.path.basename(firmware_loc)
    with open(firmware_loc, 'rb') as fle:
        output = fle.read()

    compsigname = os.path.basename(compsig_loc)
    with open(compsig_loc, 'rb') as cle:
        compsigoutput = cle.read()


    session_tuple = ('sessionKey', session_key)
    parameters_tuple = ('parameters', json.dumps(json_data))
    file_tuple = ('file', (filename, output, 'application/octet-stream'))
    compsig_tuple = ('compsig', (compsigname, compsigoutput, 'application/octet-stream'))

    #Build the payload from each multipart-form data tuple
    body.append(session_tuple)
    body.append(parameters_tuple)
    body.append(compsig_tuple)
    body.append(file_tuple)

    #Create our header dictionary
    header = {'Cookie': 'sessionKey=' + session_key}

    #We pass the whole list payload to post
    resp = _redfishobj.post(path, body, headers=header)

    if resp.status == 400:
        sys.stderr.write("Failed to upload firmware... Error: '%s'\n" % str(resp))
    elif not resp.status in [200, 201]:
        sys.stderr.write("An http response of '%s' was returned.\n" % resp.status)
    else:
        print("Upload complete!\n")

if __name__ == "__main__":
    # When running on the server locally use the following commented values
    #SYSTEM_URL = None
    #LOGIN_ACCOUNT = None
    #LOGIN_PASSWORD = None

    # When running remotely connect using the secured (https://) address,
    # account name, and password to send https requests
    # SYSTEM_URL acceptable examples:
    # "https://10.0.0.0"
    # "https://ilo.hostname"
    SYSTEM_URL = "https://10.0.0.0"
    LOGIN_ACCOUNT = "admin"
    LOGIN_PASSWORD = "password"

    # The path to the firmware component file to upload.
    # .rpm and .zip component are also valid.
    #
    FIRMWARE_PATH = "/path/to/cpxxxxxx.exe"
    COMPSIG_PATH = "/path/to//cpxxxxxx.compsig"
    # Upload the firmware file to the iLO Repository
    UPDATE_REPO = True
    # Update the system with the firmware file
    UPDATE_TARGET = False

    # flag to force disable resource directory. Resource directory and associated operations are
    # intended for HPE servers.
    DISABLE_RESOURCE_DIR = False

    try:
        # Create a Redfish client object
        REDFISHOBJ = RedfishClient(base_url=SYSTEM_URL, username=LOGIN_ACCOUNT, \
                                                                            password=LOGIN_PASSWORD)
        # Login with the Redfish client
        REDFISHOBJ.login()
    except ServerDownOrUnreachableError as excp:
        sys.stderr.write("ERROR: server not reachable or does not support RedFish.\n")
        sys.exit()

    upload_firmware(REDFISHOBJ, FIRMWARE_PATH, COMPSIG_PATH, UPDATE_REPO, UPDATE_TARGET)

    REDFISHOBJ.logout()



 # Copyright 2020 Hewlett Packard Enterprise Development LP
 #
 # Licensed under the Apache License, Version 2.0 (the "License"); you may
 # not use this file except in compliance with the License. You may obtain
 # a copy of the License at
 #
 #      http://www.apache.org/licenses/LICENSE-2.0
 #
 # Unless required by applicable law or agreed to in writing, software
 # distributed under the License is distributed on an "AS IS" BASIS, WITHOUT
 # WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. See the
 # License for the specific language governing permissions and limitations
 # under the License.

# -*- coding: utf-8 -*-
"""
An example of removing a user account
"""

import sys
import json
from redfish import RedfishClient
from redfish.rest.v1 import ServerDownOrUnreachableError

from get_resource_directory import get_resource_directory

def remove_ilo_user_account(_redfishobj, username_to_delete):

    resource_instances = get_resource_directory(_redfishobj)

    if DISABLE_RESOURCE_DIR or not resource_instances:
        #resource directory is not available so we will navigate through paths manually to obtain
        #account info
        account_service_uri = _redfishobj.root.obj['AccountService']['@odata.id']
        account_service_response = _redfishobj.get(account_service_uri)
        account_collection_uri = account_service_response.obj['Accounts']['@odata.id']
    else:
        #obtain all account instances from resource directory
        for instance in resource_instances:
            if '#ManagerAccountCollection.' in instance['@odata.type']:
                account_collection_uri = instance['@odata.id']
                break

    #find the account to delete
    account_uri_to_delete = None
    account_uris = _redfishobj.get(account_collection_uri)
    for account_uri in account_uris.dict['Members']:
        account = _redfishobj.get(account_uri['@odata.id'])
        if account.dict['UserName'] == username_to_delete:
            account_uri_to_delete = account_uri['@odata.id']
            break

    if not account_uri_to_delete:
        sys.stderr.write("Cannot find account to modify")
        return

    resp = _redfishobj.delete(account_uri_to_delete)

    if resp.status == 400:
        try:
            print(json.dumps(resp.obj['error']['@Message.ExtendedInfo'], indent=4, sort_keys=True))
        except Exception as excp:
            sys.stderr.write("A response error occurred, unable to access iLO Extended Message "\
                             "Info...")
    elif resp.status != 200:
        sys.stderr.write("An http response of \'%s\' was returned.\n" % resp.status)
    else:
        print("Success!\n")
        print(json.dumps(resp.dict, indent=4, sort_keys=True))

if __name__ == "__main__":
    # When running on the server locally use the following commented values
    #SYSTEM_URL = None
    #LOGIN_ACCOUNT = None
    #LOGIN_PASSWORD = None

    # When running remotely connect using the secured (https://) address,
    # account name, and password to send https requests
    # SYSTEM_URL acceptable examples:
    # "https://10.0.0.100"
    # "https://ilo.hostname"
    SYSTEM_URL = "https://10.0.0.100"
    LOGIN_ACCOUNT = "admin"
    LOGIN_PASSWORD = "password"

    # The username of the account to remove
    ACCOUNT_TO_DELETE = "joker"

    # flag to force disable resource directory. Resource directory and associated operations are
    # intended for HPE servers.
    DISABLE_RESOURCE_DIR = False

    try:
        # Create a Redfish client object
        REDFISHOBJ = RedfishClient(base_url=SYSTEM_URL, username=LOGIN_ACCOUNT, \
                                                                            password=LOGIN_PASSWORD)
        # Login with the Redfish client
        REDFISHOBJ.login()
    except ServerDownOrUnreachableError as excp:
        sys.stderr.write("ERROR: server not reachable or does not support RedFish.\n")
        sys.exit()

    remove_ilo_user_account(REDFISHOBJ, ACCOUNT_TO_DELETE)

    REDFISHOBJ.logout()



###
# Copyright 2020 Hewlett Packard Enterprise, Inc. All rights reserved.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#  http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
###

# -*- coding: utf-8 -*-
"""Base implementation for interaction with the iLO interface"""

# ---------Imports---------

import logging
import os
import struct
import time
from ctypes import byref, c_uint32, c_char_p, c_void_p, create_string_buffer

# ---------End of imports---------
# ---------Debug logger---------

LOGGER = logging.getLogger(__name__)


# ---------End of debug logger---------


class BlobReturnCodes(object):
    """Blob store return codes.

    SUCCESS           success

    """

    SUCCESS = 0
    if os.name != "nt":
        CHIFERR_NoDriver = 19
        CHIFERR_AccessDenied = 13
    else:
        CHIFERR_NoDriver = 2
        CHIFERR_AccessDenied = 5


class HpIloInitialError(Exception):
    """Raised when error during initialization of iLO Chif channel"""

    pass


class HpIloNoChifDriverError(Exception):
    """Raised when error during initialization of iLO Chif channel"""

    pass


class HpIloChifAccessDeniedError(Exception):
    """Raised when error during initialization of iLO Chif channel"""

    pass


class HpIloPrepareAndCreateChannelError(Exception):
    """Raised when error during initialization of iLO Chif channel"""

    pass


class HpIloChifPacketExchangeError(Exception):
    """Raised when errors encountered when exchanging chif packet"""

    pass


class HpIloReadError(Exception):
    """Raised when errors encountered when reading from iLO"""

    pass


class HpIloWriteError(Exception):
    """Raised when errors encountered when writing to iLO"""

    pass


class HpIloSendReceiveError(Exception):
    """Raised when errors encountered when reading form iLO after sending"""

    pass


class HpIloNoDriverError(Exception):
    """Raised when errors encountered when there is no ilo driver"""

    pass


class HpIlo(object):
    """Base class of interaction with iLO"""

    def __init__(self, dll=None, log_dir=None):
        fhandle = c_void_p()
        self.dll = dll
        self.log_dir = log_dir
        if LOGGER.isEnabledFor(logging.DEBUG):
            self.dll.enabledebugoutput.argtypes = [c_char_p]
            if log_dir is not None:
                logdir_c = create_string_buffer(log_dir.encode('UTF-8'))
                self.dll.enabledebugoutput(logdir_c)

        self.dll.ChifInitialize(None)

        self.dll.ChifCreate.argtypes = [c_void_p]
        self.dll.ChifCreate.restype = c_uint32

        try:
            LOGGER.debug("Calling ChifCreate...")
            status = self.dll.ChifCreate(byref(fhandle))
            if status != BlobReturnCodes.SUCCESS:
                raise HpIloInitialError("Error %s occurred while trying " "to create a channel." % status)

            self.fhandle = fhandle

            if "skip_ping" not in os.environ:
                status = self.dll.ChifPing(self.fhandle)
                if status != BlobReturnCodes.SUCCESS:
                    errmsg = "Error {0} occurred while trying to open a " "channel to iLO".format(status)
                    if status == BlobReturnCodes.CHIFERR_NoDriver:
                        errmsg = "No devices were found."
                        if os.name != "nt":
                            errmsg = "{0} Ensure the hpilo kernel module is loaded.".format(errmsg)
                    elif status == BlobReturnCodes.CHIFERR_AccessDenied:
                        errmsg = "You must be root/Administrator to use this program."
                    raise HpIloInitialError(errmsg)
                # LOGGER.debug("Calling ChifSetRecvTimeout...")
                self.dll.ChifSetRecvTimeout(self.fhandle, 60000)
        except:
            raise

    def chif_packet_exchange(self, data):
        """Function for handling chif packet exchange

        :param data: data to be sent for packet exchange
        :type data: str

        """
        datarecv = self.dll.get_max_buffer_size()
        buff = create_string_buffer(bytes(data))

        recbuff = create_string_buffer(datarecv)
        # LOGGER.debug("Calling ChifPacketExchange...")
        error = self.dll.ChifPacketExchange(self.fhandle, byref(buff), byref(recbuff), datarecv)
        if error != BlobReturnCodes.SUCCESS:
            raise HpIloChifPacketExchangeError("Error %s occurred while " "exchange chif packet" % error)

        pkt = bytearray()

        if datarecv is None:
            bytearray(recbuff[:])
        else:
            pkt = bytearray(recbuff[:datarecv])

        return pkt

    def send_receive_raw(self, data, retries=10):
        """Function implementing proper send receive retry protocol

        :param data: data to be sent for packet exchange
        :type data: str
        :param retries: number of retries for reading data from iLO
        :type retries: int

        """
        tries = 0
        sequence = struct.unpack("<H", bytes(data[2:4]))[0]

        while tries < retries:
            try:
                resp = self.chif_packet_exchange(data)

                if sequence != struct.unpack("<H", bytes(resp[2:4]))[0]:
                    if LOGGER.isEnabledFor(logging.DEBUG):
                        LOGGER.debug("Attempt %s has a bad sequence number.\n", tries + 1)
                    continue

                return resp
            except Exception as excp:
                time.sleep(1)

                if tries == (retries - 1):
                    self.close()

                    if LOGGER.isEnabledFor(logging.DEBUG) and excp:
                        LOGGER.debug("Error while reading iLO: %s", str(excp))
                    raise excp

            tries += 1

        raise HpIloSendReceiveError("iLO not responding")

    def close(self):
        """Chif close function"""
        try:
            if self.fhandle is not None:
                LOGGER.debug("Calling ChifClose...")
                self.dll.ChifClose(self.fhandle)
                self.fhandle = None
        except Exception:
            pass

    def __del__(self):
        """Chif delete function"""
        self.close()



 # Copyright 2020 Hewlett Packard Enterprise Development, LP.
 #
 # Licensed under the Apache License, Version 2.0 (the "License"); you may
 # not use this file except in compliance with the License. You may obtain
 # a copy of the License at
 #
 #      http://www.apache.org/licenses/LICENSE-2.0
 #
 # Unless required by applicable law or agreed to in writing, software
 # distributed under the License is distributed on an "AS IS" BASIS, WITHOUT
 # WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. See the
 # License for the specific language governing permissions and limitations
 # under the License.

import sys
import json
import argparse
from redfish import RedfishClient
from redfish.rest.v1 import ServerDownOrUnreachableError
global DISABLE_RESOURCE_DIR
from ilorest_util import get_resource_directory
from ilorest_util import get_gen

def reset_server(_redfishobj):
    
    managers_members_response = None
    resource_instances = get_resource_directory(_redfishobj)
    if DISABLE_RESOURCE_DIR or not resource_instances:
        #if we do not have a resource directory or want to force it's non use to find the
        #relevant URI
        managers_uri = _redfishobj.root.obj['Systems']['@odata.id']
        managers_response = _redfishobj.get(managers_uri)
        managers_members_uri = next(iter(managers_response.obj['Members']))['@odata.id']
        managers_members_response = _redfishobj.get(managers_members_uri)
    else:
        #Use Resource directory to find the relevant URI
        for instance in resource_instances:
            if "ComputerSystem." in instance['@odata.type']:
                managers_members_uri = instance['@odata.id']
                managers_members_response = _redfishobj.get(managers_members_uri)
                
    if managers_members_response:
        path = managers_members_response.obj["Actions"]["#ComputerSystem.Reset"]["target"]
        body = dict()
        resettype = ['ForceRestart','GracefulRestart']
        body["Action"] = "ComputerSystem.Reset"
        for reset in resettype:
            if reset.lower() == "forcerestart":
                body['ResetType'] = "ForceRestart"
                resp = _redfishobj.post(path, body)
            elif reset.lower() == "gracefulrestart":
                body['ResetType'] = "GracefulRestart"
                resp = _redfishobj.post(path, body)

    #If iLO responds with soemthing outside of 200 or 201 then lets check the iLO extended info
    #error message to see what went wrong
    if resp.status == 400:
        try:
            print(json.dumps(resp.obj['error']['@Message.ExtendedInfo'], indent=4, sort_keys=True))
        except Exception as excp:
            sys.stderr.write("A response error occurred, unable to access iLO Extended Message "\
                             "Info...")
    elif resp.status != 200:
        sys.stderr.write("An http response of \'%s\' was returned.\n" % resp.status)
    else:
        print("Success!\n")
        print(json.dumps(resp.dict, indent=4, sort_keys=True))
        
def reset_server_gen9(_redfishobj):
    
    managers_uri = "/redfish/v1/Systems/1/"
    managers_response = _redfishobj.get(managers_uri)
    system_path = managers_response.obj["Actions"]["#ComputerSystem.Reset"]["target"]
    print(system_path)
    body = dict()
    body["Action"] = "Reset"
    body["ResetType"] = "ForceRestart"

    resp = _redfishobj.post(system_path, body)
    #If iLO responds with soemthing outside of 200 or 201 then lets check the iLO extended info
    #error message to see what went wrong
    if resp.status == 400:
        try:
            print(json.dumps(resp.obj['error']['@Message.ExtendedInfo'], indent=4, sort_keys=True))
        except Exception as excp:
            sys.stderr.write("A response error occurred, unable to access iLO Extended Message "\
                             "Info...")
    elif resp.status != 200:
        sys.stderr.write("An http response of \'%s\' was returned.\n" % resp.status)
    else:
        print("Success!\n")
        print(json.dumps(resp.dict, indent=4, sort_keys=True))

if __name__ == "__main__":
    # Initialize parser 
    parser = argparse.ArgumentParser(description = "Script to upload and flash NVMe FW")    

    parser.add_argument(
        '-i',
        '--ilo',
        dest='ilo_ip',
        action="store",
        help="iLO IP of the server",
        default=None)
    parser.add_argument(
        '-u',
        '--user',
        dest='ilo_user',
        action="store",
        help="iLO username to login",
        default=None)
    parser.add_argument(
        '-p',
        '--password',
        dest='ilo_pass',
        action="store",
        help="iLO password to log in.",
        default=None)
    
    options = parser.parse_args()

    system_url = "https://" + options.ilo_ip
    print (system_url)

    # flag to force disable resource directory. Resource directory and associated operations are
    # intended for HPE servers.
    DISABLE_RESOURCE_DIR = True

    try:
        # Create a Redfish client object
        redfish_obj = RedfishClient(base_url=system_url, username=options.ilo_user, password=options.ilo_pass)
        # Login with the Redfish client
        redfish_obj.login()
    except ServerDownOrUnreachableError as excp:
        sys.stderr.write("ERROR: server not reachable or does not support RedFish.\n")
        sys.exit()
    
    (ilogen,_) = get_gen(redfish_obj)
    print ("Generation is ", ilogen)
    if int(ilogen) == 5:
        reset_server(redfish_obj)
    else:
        reset_server_gen9(redfish_obj)
    redfish_obj.logout()



from rmc import RmcApp


class Rmc_API(RmcApp):
    def __init__(self, Args=None):
        RmcApp.__init__(self, Args)

    def Api_login(self, url=None, username=None, password=None):
        self.login(base_url=url, username=username, password=password)


# for testing
if __name__ == "__main__":
    api = Rmc_API()
    api.do_stuff()



 # Copyright 2020 Hewlett Packard Enterprise Development LP
 #
 # Licensed under the Apache License, Version 2.0 (the "License"); you may
 # not use this file except in compliance with the License. You may obtain
 # a copy of the License at
 #
 #      http://www.apache.org/licenses/LICENSE-2.0
 #
 # Unless required by applicable law or agreed to in writing, software
 # distributed under the License is distributed on an "AS IS" BASIS, WITHOUT
 # WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. See the
 # License for the specific language governing permissions and limitations
 # under the License.

# -*- coding: utf-8 -*-
"""
An example of adding a user account by iLO privileges or redfish standard roles
"""

import sys
import json
from redfish import RedfishClient
from redfish.rest.v1 import ServerDownOrUnreachableError
global DISABLE_RESOURCE_DIR

from ilorest_util import get_resource_directory
from ilorest_util import get_gen

def add_ilo_user_account(_redfishobj, new_loginname, new_username, new_password, role_id, \
                         privilege_dict):
    resource_instances = get_resource_directory(_redfishobj)

    if DISABLE_RESOURCE_DIR or not resource_instances:
        #resource directory is not available so we will navigate through paths manually to obtain
        #account info
        account_service_uri = _redfishobj.root.obj['AccountService']['@odata.id']
        account_service_response = _redfishobj.get(account_service_uri)
        account_collection_uri = account_service_response.obj['Accounts']['@odata.id']
        #Add via role id
        body = {"RoleId": role_id}
    else:
        #obtain all account instances from resource directory
        for instance in resource_instances:
            if '#ManagerAccountCollection.' in instance['@odata.type']:
                account_collection_uri = instance['@odata.id']
        body = {"Oem": {"Hpe": {"Privileges": {}}}}
        #HPE server, so add via privileges
        for priv in privilege_dict:
            body["Oem"]["Hpe"]["Privileges"][priv] = privilege_dict[priv]
        #Add login name
        body["Oem"]["Hpe"]["LoginName"] = new_loginname

    #Fill in the rest of the payload
    body["UserName"] = new_username
    body["Password"] = new_password

    #We pass the URI and the dictionary as a POST command (part of the redfish object)
    resp = _redfishobj.post(account_collection_uri, body)

    #If iLO responds with soemthing outside of 200 or 201 then lets check the iLO extended info
    #error message to see what went wrong
    if resp.status == 400:
        try:
            print(json.dumps(resp.obj['error']['@Message.ExtendedInfo'], indent=4, sort_keys=True))
        except Exception:
            sys.stderr.write("A response error occurred, unable to access iLO Extended Message "\
                             "Info...")
    elif not resp.status in [200, 201]:
        sys.stderr.write("An http response of '%s' was returned.\n" % resp.status)
    else:
        print("Success!\n")
        print(json.dumps(resp.dict, indent=4, sort_keys=True))

def add_ilo_user_account_gen9(_redfishobj, new_loginname, new_username, new_password):
	
	account_collection_uri = "/redfish/v1/AccountService/Accounts/"
	#Add via gen9 priv dic
	body = {'Oem': {'Hp': {'Privileges': {"LoginPriv": True, "RemoteConsolePriv": True,
        "UserConfigPriv": True, "VirtualMediaPriv": True, "VirtualPowerAndResetPriv": True,
        "iLOConfigPriv": True}, 'LoginName': new_loginname}},'UserName': new_username, 'Password': new_password}
	#We pass the URI and the dictionary as a POST command (part of the redfish object)
	resp = _redfishobj.post(account_collection_uri, body)
	print(json.dumps(resp.dict, indent=4, sort_keys=True))

if __name__ == "__main__":

    # When running on the server locally use the following commented values
    #SYSTEM_URL = None
    #LOGIN_ACCOUNT = None
    #LOGIN_PASSWORD = None

    # When running remotely connect using the secured (https://) address,
    # account name, and password to send https requests
    # SYSTEM_URL acceptable examples:
    # "https://10.0.0.0"
    # "https://ilo.hostname"
    if len(sys.argv) == 4:
        # Remote mode
        SYSTEM_URL = sys.argv[1]
        LOGIN_ACCOUNT = sys.argv[2]
        LOGIN_PASSWORD = sys.argv[3]
    else:
        # Local mode
        SYSTEM_URL = None
        LOGIN_ACCOUNT = None
        LOGIN_PASSWORD = None

    #account login name (iLO GUI actually considers this to be 'UserName', but
    #this is the redfish standard username)
    ACCOUNT_LOGIN_NAME = "batman"

    #account user name (iLO GUI actually considers this to be 'LoginName', but
    #this is the redfish login)
    ACCOUNT_USER_NAME = "bruce_wayne"

    #account password
    ACCOUNT_PASSWORD = "thedarkknight123"

    #A predefined role for the user, (The redfish standard method for accounts).
    #This is a translated to a pre-configured arrangement of privileges on HPE servers
    ROLE_ID = "Administrator" #Administrator, ReadOnly or Operator are available

    #Dictionary of modifiable privileges for HPE servers (modify this if you wish to directly set
    #an account with specific privileges
    PRIVILEGE_DICT = {"iLOConfigPriv": True, "VirtualMediaPriv": True, "RemoteConsolePriv": True,\
                      "UserConfigPriv": True, "VirtualPowerAndResetPriv": True, \
                      "SystemRecoveryConfigPriv": False, "LoginPriv": True, \
                      "HostStorageConfigPriv": True, "HostNICConfigPriv": True, \
                      "HostBIOSConfigPriv": True}
    # flag to force disable resource directory. Resource directory and associated operations are
    # intended for HPE servers.
    DISABLE_RESOURCE_DIR = False

	#For certificate login
    #ca_cert_data = {}
    #ca_cert_data["cert_file"] = "c:\\test\\ppcacuser.crt"
    #ca_cert_data["key_file"] = "c:\\test\\ppcacuserpriv.key"
    #ca_cert_data["key_password"] = "password"
    #LOGIN_ACCOUNT = None
    #LOGIN_PASSWORD = None

    try:
        # Create a Redfish client object
		
		REDFISHOBJ = RedfishClient(base_url=SYSTEM_URL, username=LOGIN_ACCOUNT, password=LOGIN_PASSWORD)
		REDFISHOBJ.login()
		
		# For Certificate login
        #REDFISHOBJ = RedfishClient(base_url=SYSTEM_URL, username=LOGIN_ACCOUNT, password=LOGIN_PASSWORD, ca_cert_data=ca_cert_data)
        #REDFISHOBJ = RedfishClient(base_url=SYSTEM_URL, ca_cert_data=ca_cert_data)
        # Login with the Redfish client
		
		
        #if ca_cert_data is None:
        #    REDFISHOBJ.login()
        #else:
        #    REDFISHOBJ.login(auth='certificate')
    except ServerDownOrUnreachableError as excp:
        sys.stderr.write("ERROR: server not reachable or does not support RedFish.\n")
        sys.exit()

    #obtain all account instances, by navigating set paths and keys to find the relevant URI
    #(account_collection_uri, accounts, rd) = get_accounts(redfishobj, DISABLE_RESOURCE_DIR)
    #print("\n\nShowing accounts before changes:\n\n")
    #show_accounts(redfishobj, accounts)

    #if account_collection_uri and accounts:
    #add specified account
    (ilogen,_) = get_gen(REDFISHOBJ)
    print ("Generation is ", ilogen)
    if int(ilogen) >= 5:
        add_ilo_user_account(REDFISHOBJ,ACCOUNT_LOGIN_NAME,ACCOUNT_USER_NAME,ACCOUNT_PASSWORD,ROLE_ID,PRIVILEGE_DICT)
    else:
        add_ilo_user_account_gen9(REDFISHOBJ,ACCOUNT_LOGIN_NAME,ACCOUNT_USER_NAME,ACCOUNT_PASSWORD)
    REDFISHOBJ.logout()



###
# Copyright 2020 Hewlett Packard Enterprise, Inc. All rights reserved.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#  http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
###
# -*- coding: utf-8 -*-
"""Utility functions for internal and external use. Contains general json navigating functions as
well as some monolith utility functions."""
import copy
import logging
import re
import sys

import six

if six.PY3:
    from functools import reduce

try:
    from collections import Mapping
except ImportError:
    from collections.abc import Mapping

import jsonpath_rw
from six import iterkeys, string_types

from redfish.ris.rmc_helper import IncorrectPropValue

try:
    # itertools ifilter compatibility for python 2
    from future_builtins import filter
except ImportError:
    # filter function provides the same functionality in python 3
    pass

# ---------Debug logger---------

LOGGER = logging.getLogger()


# ---------End of debug logger---------


def print_handler(msg):
    """Helper function for handling warning messages appropriately. If LOGGER level is set to 40
    print out the warnings, else log them as a warning.

    :param msg: The warning message.
    :type msg: str
    """
    # if override:
    sys.stdout.write(msg)
    # else:
    # if LOGGER.getEffectiveLevel() == 40:
    # LOGGER.warning(msg)


def warning_handler(msg, override=False):
    """Helper function for handling warning messages appropriately. If LOGGER level is set to 40
    print out the warnings, else log them as a warning.

    :param msg: The warning message.
    :type msg: str
    """
    # if override:
    # sys.stdout.write(msg)
    # else:
    # if LOGGER.getEffectiveLevel() == 40:
    LOGGER.warning(msg)


def validate_headers(instance, verbose=False):
    """Validates an instance is patchable.

    :param instance: Instance of the property to check.
    :type instance: :class:`redfish.ris.RisMonolithMemberv100`
    :param verbose: Flag to print or log more information.
    :type verbose: bool
    :returns: True if the setting is not patchable, False if it is.
    """
    skip = False
    try:
        headervals = instance.resp.getheaders()
        for kii, val in headervals.items():
            if kii.lower() == "allow":
                if "PATCH" not in val:
                    if verbose:
                        warning_handler("Skipping read-only path: %s\n" % instance.resp.request.path)
                    skip = True
    except:
        pass
    return skip


def merge_dict(currdict, newdict):
    """Merges dictionaries together.

    :param currdict: Dictionary that will absorb the second.
    :type currdict: dict
    :param newdict: Dictionary to merge into the first.
    :type newdict: dict
    """
    for k, itemv2 in list(newdict.items()):
        itemv1 = currdict.get(k)

        if isinstance(itemv1, Mapping) and isinstance(itemv2, Mapping):
            merge_dict(itemv1, itemv2)
        elif isinstance(itemv1, list) and isinstance(itemv2, list):
            j = 0
            for i in itemv2:
                if isinstance(i, dict):
                    for k, v in i.items():
                        if k not in [d.keys() for d in itemv1]:
                            itemv1[j][k] = v
                            j = j + 1
                else:
                    itemv1.append(i)
        else:
            currdict[k] = itemv2


def get_errmsg_type(results):
    """Return the registry type of a response.

    :param results: rest response.
    :type results: :class:`redfish.rest.containers.RestResponse`
    :returns: A Registry Id type string, None if not match is found, or no_id if the
              response is not an error message
    :rtype: None or string
    """

    message_type = None
    try:
        jsonpath_expr = jsonpath_rw.parse("$..MessageId")
        messageid = [match.value for match in jsonpath_expr.find(results.dict)]
        if not messageid:
            jsonpath_expr = jsonpath_rw.parse("$..MessageID")
            messageid = [match.value for match in jsonpath_expr.find(results.dict)]
        if messageid:
            message_type = messageid[0].split(".")[0]
    except:
        pass

    return message_type


def filter_output(output, sel, val):
    """Filters a list of dictionaries based on a key:value pair only returning the dictionaries
    that include the key and value.

    :param output: List of dictionaries to check for the key:value.
    :type output: list
    :param sel: the key for the property to be filtered by.
    :type sel: str
    :param val: value for the property be filtered by.
    :type val: str
    :returns: A filtered list from output parameter
    :rtype: list
    """
    # TODO: check if this can be replaced by navigatejson
    newoutput = []
    if isinstance(output, list):
        for entry in output:
            if isinstance(entry, dict):
                if "/" in sel:
                    sellist = sel.split("/")
                    newentry = copy.copy(entry)

                    for item in sellist:
                        if item in list(newentry.keys()):
                            if item == sellist[-1] and str(newentry[item]) == str(val):
                                newoutput.append(entry)
                            else:
                                newentry = newentry[item]
                else:
                    if sel in list(entry.keys()) and entry[sel] == val:
                        newoutput.append(entry)
            else:
                return output

    return newoutput


def checkallowablevalues(newdict=None, oridict=None):
    """Validate dictionary changes with Redfish allowable values. This will raise an
    :class:`redfish.ris.rmc_helper.IncorrectPropValue` error if the dictionary is not valid.

    :param newdict: dictionary with only the properties that have changed.
    :type newdict: dict
    :param oridict: Full dictionary with current state. (Includes @Redfish.AllowableValues)
    :type oridict: dict
    """
    for strmatch in re.finditer("@Redfish.AllowableValues", str(oridict)):
        propname = str(oridict)[: strmatch.start()].split("'")[-1]
        strtomatch = "$..'{0}@Redfish.AllowableValues'".format(propname)
        jsonpath_expr = jsonpath_rw.parse(strtomatch)
        matches = jsonpath_expr.find(oridict)
        if matches:
            for match in matches:
                fullpath = str(match.full_path)
                if "Actions" in fullpath:
                    continue
                checkpath = fullpath.split("@Redfish.AllowableValues")[0]
                jexpr2 = jsonpath_rw.parse(checkpath)
                valmatches = jexpr2.find(newdict)
                if valmatches:
                    for mat in valmatches:
                        res = [val for val in match.value if mat.value.lower() == val.lower()]
                        if not res:
                            raise IncorrectPropValue(
                                "Incorrect Value "
                                "entered. Please enter one of the below "
                                "values for {0}:\n{1}".format("/".join(checkpath.split(".")), str(match.value)[1:-1])
                            )


def navigatejson(selector, currdict, val=None):
    """Function for navigating the json dictionary. Searches a dictionary for specific keys
    and possibly values, returning only the dictionary sections for the requested keys and values.

    :param selector: the property required from current dictionary.
    :type selector: list
    :param val: value to be filtered by.
    :type val: str or int or bool
    :param currdict: json dictionary of list to be filtered
    :type currdict: json dictionary/list
    :returns: returns a dictionary of selected items
    """
    # TODO: Check for val of different types(bool, int, etc)
    temp_dict = dict()
    createdict = lambda y, x: {x: y}
    getkey = lambda cdict, sel: next((item for item in iterkeys(cdict) if sel.lower() == item.lower()), sel)
    getval = lambda cdict, sele: [cdict[sel] if sel in cdict else "~!@#$%^&*)()" for sel in [getkey(cdict, sele)]][0]
    fullbreak = False
    seldict = copy.deepcopy(currdict)
    for ind, sel in enumerate(selector):
        if isinstance(seldict, dict):
            selector[ind] = getkey(seldict, sel)
            seldict = getval(seldict, sel)
            if seldict == "~!@#$%^&*)()":
                return None
            if val and ind == len(selector) - 1:
                cval = ",".join(seldict) if isinstance(seldict, (list, tuple)) else seldict
                if not (
                    (val[-1] == "*" and str(cval).lower().startswith(val[:-1].lower()))
                    or str(cval).lower() == val.lower()
                ):
                    fullbreak = True
        elif isinstance(seldict, (list, tuple)):
            returndict = []
            for items in seldict:
                correctcase = selector[ind:]
                returnseldict = navigatejson(correctcase, items)
                selector[ind:] = correctcase
                if returnseldict is not None:
                    returndict.append(returnseldict)
            if returndict:
                seldict = returndict
            else:
                fullbreak = True
            if seldict:
                seldict = {selector[ind - 1]: seldict}
                selsdict = reduce(createdict, [seldict] + selector[: ind - 1][::-1])
                merge_dict(temp_dict, selsdict)
                return temp_dict
            else:
                break
        else:
            fullbreak = True
            break
    if fullbreak:
        return None
    else:
        selsdict = reduce(createdict, [seldict] + selector[::-1])
        merge_dict(temp_dict, selsdict)
    return temp_dict


def iterateandclear(dictbody, proplist):
    """Iterate over a dictionary and remove listed properties.

    :param dictbody: json body
    :type dictbody: dict or list
    :param proplist: property list
    :type proplist: list
    """
    if isinstance(dictbody, dict):
        _ = [dictbody.pop(key) for key in proplist if key in dictbody]
        for key in dictbody:
            dictbody[key] = iterateandclear(dictbody[key], proplist)
    if isinstance(dictbody, list):
        for ind, val in enumerate(dictbody):
            dictbody[ind] = iterateandclear(val, proplist)
    return dictbody


def skipnonsettingsinst(instances):
    """Removes non /settings sections. Useful for only returning settings monolith members.
    Example: Members with paths `/redfish/v1/systems/1/bios/` and
    `/redfish/v1/systems/1/bios/settings`
    will return only the `/redfish/v1/systems/1/bios/settings` member.

    :param instances: list of :class:`redfish.ris.ris.RisMonolithMemberv100`
      instances to check for settings paths.
    :type instances: list
    :returns: list of :class:`redfish.ris.ris.RisMonolithMemberv100` setting instances
    :rtype: list
    """
    instpaths = [inst.path.lower() for inst in instances]
    cond = list(filter(lambda x: x.endswith(("/settings", "settings/")), instpaths))
    paths = [path.split("settings/")[0].split("/settings")[0] for path in cond]
    newinst = [inst for inst in instances if inst.path.lower() not in paths]
    return newinst


def getattributeregistry(instances, adict=None):
    # add try except return {} after test
    """Gets an attribute registry in given monolith instances.

    :param instances: list of :class:`redfish.ris.ris.RisMonolithMemberv100` instances to be
      checked for attribute registry.
    :type instances: list
    :param adict: A dictionary containing an AttributeRegistry
    :type adict: dict
    :return: returns a dictionary containing the attribute registry string(s)
    :rtype: dict
    """

    if adict:
        return adict.get("AttributeRegistry", None)
    newdict = {}
    for inst in instances:
        try:
            if "AttributeRegistry" in inst.resp.dict:
                if inst.defpath is not None:
                    if not ("bios/settings" in inst.defpath):
                        newdict[inst.maj_type] = inst.resp.obj["AttributeRegistry"]
                        return newdict
                newdict[inst.maj_type] = inst.resp.obj["AttributeRegistry"]
        except AttributeError:
            LOGGER.warning("Invalid/Unpopulated Response: %s\nType:%s\nPath:%s\n" % (inst.resp, inst.type, inst.path))
    return newdict


def diffdict(newdict=None, oridict=None, settingskipped=[False]):
    """Diff two dictionaries, returning only the values that are different between the two.

    :param newdict: The first dictionary to check for differences.
    :type newdict: dict
    :param oridict: The second dictionary to check for differences.
    :type oridict: dict
    :param settingskipped: Flag to determine if any settings were missing.
    :type settingskipped: list
    :returns: dictionary with only the properties that have changed.
    :rtype: dict
    """
    try:
        if newdict == oridict:
            return {}
    except:
        try:
            if set(newdict) == set(oridict):
                return {}
        except:
            pass

    newdictkeys = list(newdict.keys())
    newdictlist = []
    if type(oridict) is list:
        oridict = oridict[0]
        newdictlist.append(newdict)
    oridictkeys = list(oridict.keys())
    newdictkeyslower = [ki.lower() for ki in newdictkeys]
    oridictkeyslower = [ki.lower() for ki in list(oridict.keys())]
    missingkeys = list(set(newdictkeyslower) - set(oridictkeyslower))
    for kis in missingkeys:
        del newdict[newdictkeys[newdictkeyslower.index(kis)]]
        warning_handler("Attribute {0} not found in the selection...".format(kis))
        settingskipped = [True]
    for key, val in list(newdict.items()):
        if key not in oridict:
            keycase = oridictkeys[oridictkeyslower.index(key.lower())]
            del newdict[key]
            key = keycase
            newdict[key] = val
        if isinstance(val, dict):
            res = diffdict(newdict[key], oridict[key])
            if res:
                newdict[key] = res
            else:
                del newdict[key]
        elif isinstance(val, list):
            if val == oridict[key]:
                del newdict[key]
                continue
            if len(val) == 1 and isinstance(val[0], dict):
                if newdict[key] and oridict[key]:
                    res = diffdict(newdict[key][0], oridict[key][0], settingskipped)
                    if res:
                        newdict[key][0] = res
                    else:
                        del newdict[key]
            if [li for li in val if not isinstance(li, string_types)]:
                continue
            else:
                if val:
                    if [va.lower() for va in val] == [va.lower() if va else va for va in oridict[key]]:
                        del newdict[key]
        # TODO: check if lowercase is correct or buggy for string types
        elif isinstance(val, (string_types, int, type(None))):
            if newdict[key] == oridict[key]:
                del newdict[key]
    if not newdictlist:
        return newdict
    else:
        return newdictlist


def json_traversal(data, key_to_find, ret_dict=False):
    """
    PENDING MODIFICATION TO MORE GENERALIZED NOTATION

    Recursive function to traverse a JSON resposne object and retrieve the array of
    relevant data (value or full key/value pair). Only a single key needs to be found within the
    dictionary in order to return a valid dictionary or value.

    #Intended Usage:
    - Error response message parsing
    - Checkreadunique in Validation

    :param data: json data to be parsed
    :type data: JSON error response object
    :param key_to_find: JSON key to be found
    :type key_to_find: String
    :param ret_dict: return dictionary instead of just value
    :type ret_dict: boolean
    :returns: value or dictionary containing 'key_to_find'
                (and all additional keys at the same level).
    """

    try:
        for i, _iter in enumerate(data):
            try:
                if _iter == data:
                    return None
            except Exception:
                pass
            try:
                if key_to_find.lower() == _iter.lower():
                    if ret_dict:
                        return data
                    else:
                        return data[_iter]
            except Exception:
                pass
            try:
                if key_to_find.lower() in [str(_.lower()) for _ in _iter.keys()]:
                    if ret_dict:
                        return data
                    else:
                        return data[_iter]
            except Exception:
                pass
            _tmp = None
            try:
                if isinstance(data[_iter], dict):
                    for k in data[_iter].keys():
                        if k.lower() == key_to_find.lower():
                            if ret_dict:
                                return data[_iter]
                            else:
                                return data[_iter][k]
                    _tmp = json_traversal(data[_iter], key_to_find, ret_dict)
                elif isinstance(data[_iter], list) or isinstance(data[_iter], tuple):
                    try:
                        _tmp = json_traversal(data[i], key_to_find, ret_dict)
                    except Exception:
                        _tmp = json_traversal(data[_iter], key_to_find, ret_dict)
            except Exception:
                _tmp = json_traversal(data[i], key_to_find, ret_dict)
            finally:
                if _tmp:
                    return _tmp
    except Exception:
        pass


def json_traversal_delete_empty(data, old_key=None, _iter=None, remove_list=None):
    """
    Recursive function to traverse a dictionary and delete things which
    match elements in the remove_list

    :param data: to be truncated
    :type data: list or dict
    :param old_key: key from previous recursive call (higher in stack)
    :type old_key: dictionary key
    :param _iter: iterator tracker for list (tracks iteration across
    recursive calls)
    :type _iter: numerical iterator
    :param remove_list: list of items to be removed
    :type: list
    :returns: none
    """

    if not remove_list:
        remove_list = ["NONE", None, "", {}, [], "::", "0.0.0.0", "Unknown"]
    list_quick_scan = False

    if isinstance(data, list):
        if _iter is None:
            for idx, val in enumerate(data):
                if idx is (len(data) - 1):
                    list_quick_scan = True

                json_traversal_delete_empty(val, old_key, idx, remove_list)

            if list_quick_scan:
                for j in remove_list:
                    for _ in range(data.count(j)):
                        data.remove(j)

    elif isinstance(data, dict):
        delete_list = []
        for key, value in data.items():
            if (
                (isinstance(value, dict) and len(value) < 1)
                or (isinstance(value, list) and len(value) < 1)
                or None
                or value in remove_list
                or key in remove_list
            ):
                delete_list.append(key)

            else:
                json_traversal_delete_empty(value, key, remove_list=remove_list)
                # would be great to not need this section; however,
                # since recursive deletion is not possible, this is needed
                # if you can figure out how to pass by reference then fix me!
                if (isinstance(value, dict) and len(value) < 1) or None or value in remove_list:
                    delete_list.append(key)
        for dl_entry in delete_list:
            try:
                del data[dl_entry]
            except KeyError:
                pass



###
# Copyright 2020 Hewlett Packard Enterprise, Inc. All rights reserved.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#  http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
###
# -*- coding: utf-8 -*-

"""Error response handler for Redfish or LegacryRest responses. Extended information only available
with registries available on system, otherwise will return generic error responses."""
import logging

from redfish.ris.ris import SessionExpired
from redfish.ris.rmc_helper import (
    EmptyRaiseForEAFP,
    IdTokenError,
    IloLicenseError,
    IloResponseError,
    ScepenabledError,
    ValueChangedError,
)
from redfish.ris.utils import (
    get_errmsg_type,
    json_traversal,
    print_handler,
    warning_handler,
)

# ---------Debug logger---------

LOGGER = logging.getLogger()


# ---------End of debug logger---------


class ResponseHandler(object):
    """Class to handle error responses from the server.

    :param validation_mgr: ValidationManager instance to gather registries if needed. Available
                           in an RmcApp class as an attribute.
    :type validation_mgr: ValidationManager
    :param msg_reg_type: Redfish (#MessageRegistry.) or LegacyRest (MessageRegistry.)
                         message registry string. Available in Typesandpathdefines class.
    :type msg_reg_type: str
    """

    def __init__(self, validaition_mgr, msg_type):
        self.validation_mgr = validaition_mgr
        self.msg_reg_type = msg_type

    def output_resp(self, response, dl_reg=False, verbosity=1):
        """Prints or logs parsed MessageId response. Will raise an IloResponseError or return
        a list of message response data which includes the information returned from
        message_handler.

        :param response: message response of a call.
        :type response: :class:`redfish.rest.containers.RestResponse`
        :param dl_reg: Flag to download registry. If this is set to True a generic message response
                       will be returned instead of gathered from registries.
        :type dl_reg: bool
        :param verbosity: Optional verbosity level. Only modifies what is output to log or screen.
        :type verbosity: int
        :returns: List of error response dictionaries.
        """
        retdata = None

        if response.status > 299:
            message_text = "No error message returned or unable to parse error response."
        else:
            message_text = "The operation completed successfully."

        if response.status < 300 and (response._rest_request.method == "GET" or not response.read):
            # for rawget
            print_handler(
                self.verbosity_levels(
                    message=message_text, response_status=response.status, verbosity=verbosity, dl_reg=dl_reg
                )
            )
        elif response.status == 401:
            raise SessionExpired()
        elif response.status == 403:
            results = response.dict["error"]["@Message.ExtendedInfo"]
            for result in results:
                if "License" in list(result.values())[0] or "license" in list(result.values())[0]:
                    raise IloLicenseError("")
            raise IdTokenError()
        elif response.status == 412:
            warning_handler(
                "The property you are trying to change has been updated. "
                "Please check entry again before manipulating it.\n",
                override=True,
            )
            raise ValueChangedError()
        else:
            retdata = self.message_handler(
                response_data=response,
                verbosity=verbosity,
                message_text=message_text,
                dl_reg=dl_reg,
            )
        if response.status == 400:
            results = response.dict["error"]["@Message.ExtendedInfo"]
            for result in results:
                if "License" in list(result.values())[0] or "license" in list(result.values())[0]:
                    raise IloLicenseError("")
                if "UnsupportedOperationACEEnabled" in list(result.values())[0]:
                    raise ScepenabledError("")
        if response.status > 299:
            raise IloResponseError("")
        else:
            return retdata

    def message_handler(self, response_data, verbosity=0, message_text="No Response", dl_reg=False):
        """Prints or logs parsed MessageId response based on verbosity level and returns the
        following message information in a list:

        * MessageArgs
        * MessageId
        * RestResponse status
        * Resolution
        * Full error message text

        :param response_data: message response of a call.
        :type response_data: :class:`redfish.rest.containers.RestResponse`
        :param verbosity: Optional verbosity level. Only modifies what is output to log or screen.
        :type verbosity: int
        :param message_text: Response message text. If not provided, message_handler will attempt to
                             parse it from the RestResponse and registries.
        :type message_text: str
        :param dl_reg: Flag to download registry. If this is set to True a generic message response
                       will be returned instead of gathered from registries.
        :type dl_reg: bool
        :returns: List of error response dictionaries.
        """
        _tmp_message_id = _tmp_description = _tmp_resolution = message_text
        retlist = list()
        response_error_str = ""
        try:
            response_status = response_data.status
        except (AttributeError, ValueError):
            response_status = "???"
        try:
            response_data = response_data.dict
        except (AttributeError, ValueError):
            pass
        try:
            for inst in self.get_message_data(response_data, dl_reg):
                try:
                    for _key in inst.keys():
                        if "messageid" in str(_key.lower()):
                            _tmp_message_id = inst[_key]
                        if "description" in str(_key.lower()):
                            _tmp_description = inst[_key]
                    if inst.get("Message") and inst.get("MessageArgs"):
                        for i in range(inst["Message"].count("%")):
                            inst["Message"] = inst["Message"].replace(
                                "%" + str(i + 1), '"' + inst["MessageArgs"][i] + '"'
                            )
                        message_text = inst.get("Message", " ")
                    elif inst.get("Message"):
                        message_text = inst.get("Message", " ")
                    elif response_status not in [200, 201]:
                        message_text = _tmp_message_id
                    _tmp_resolution = inst.get("Resolution", " ")
                except (KeyError, ValueError, TypeError):
                    pass
                finally:
                    response_error_str += "[%s] %s\n" % (response_status, message_text)
                    print_handler(
                        self.verbosity_levels(
                            message_text,
                            _tmp_message_id,
                            _tmp_description,
                            _tmp_resolution,
                            response_status,
                            verbosity,
                            dl_reg,
                        )
                    )
                    retlist.append(inst)
        except Exception:
            if not message_text:
                message_text = _tmp_message_id
            response_error_str += "[%s] %s\n" % (response_status, message_text)
            print_handler(
                self.verbosity_levels(
                    message_text,
                    _tmp_message_id,
                    _tmp_description,
                    _tmp_resolution,
                    response_status,
                    verbosity,
                    dl_reg,
                )
            )
            retlist.append(inst)
        finally:
            return retlist

    def get_message_data(self, resp_data, dl_reg=False):
        """Obtain relevant keys from rest response.

        :param resp: response
        :type resp: :class:`redfish.rest.containers.RestResponse`
        :returns: list of error response dictionaries
        """
        err_response_keys = ["MessageId", "Message", "MessageArgs", "Resolution"]
        try:
            if "messageid" in [_key.lower() for _key in resp_data.keys()]:
                data_extract = [resp_data]
            else:
                raise TypeError
        except (TypeError, KeyError):
            data_extract = json_traversal(resp_data, "messageid", ret_dict=True)
        if data_extract:
            try:
                if not dl_reg:
                    for inst in data_extract:
                        if [key.lower() for key in inst.keys()] not in [erk.lower() for erk in err_response_keys]:
                            if "messageid" in [str(key.lower()) for key in inst.keys()]:
                                inst.update(self.get_error_messages(inst[key]))
                                continue
            finally:
                return data_extract
        else:
            return None

    def verbosity_levels(
        self,
        message,
        messageid=" ",
        description=" ",
        resolution=" ",
        response_status=None,
        verbosity=0,
        dl_reg=False,
    ):
        """Formatting based on verbosity level.

        :param message: Message from BMC response combined with the registry model/schema.
        :type message: str
        :param messageid: Error code as classified by the BMC's error code registry.
        :type messageid: str
        :param resolution: Message from BMC registry model/schema with the suggested
                           resolution for the given error.
        :type resolution: str
        :param response_status: HTTP response status code.
        :type response_status: int
        :param verbosity: Option to set/control output message (stderr) verbosity.
        :type verbosity: int
        :returns: Message to be returned to caller.
        """
        resp_str = ""
        if response_status:
            resp_str = "[" + str(response_status) + "] "

        if (verbosity == 1 or dl_reg) and message:
            return resp_str + message + "\n"
        elif verbosity > 1 and messageid and message and resolution:
            if not resp_str:
                resp_str = "None "
            return (
                "\nHTTP Response Code: "
                + resp_str[:-1]
                + "\nMessageId: "
                + messageid
                + "\nDescription: "
                + description
                + "\nMessage: "
                + message
                + "\nResolution: "
                + resolution
                + "\n"
            )
        else:
            if response_status == 400:
                return "" + message + "\n"
            else:
                return ""

    # unused? (removal pending)
    @staticmethod
    def _get_errmsg_type(results):
        """Return the registry type of a response.

        :param resuts: rest response.
        :type results: RestResponse.
        :returns: returns a Registry Id type string, None if not match is found, or no_id if the
                  response is not an error message.
        """
        return get_errmsg_type(results)

    def get_error_messages(self, regtype=None):
        """Returns registry error messages. Can specify a specific registry to return by Id.

        :param regtype: Id of registry type to add to list.
        :type regtype: str
        :returns: A list of error messages.
        """
        LOGGER.info("Entering validation...")
        messages = None
        errmessages = {}
        reglist = []

        # An error occurred during the shortcut method so let's go through each registry,
        # obtain the schema and narrow down the selected schema for the registry type provided
        try:
            _regtype = regtype.split(".")[0]
            for reg in self.validation_mgr.iterregmems():
                # gen 10 / gen 9 rest
                if _regtype:
                    if reg and "Id" in reg and reg["Id"] == _regtype:
                        try:
                            reglist.append(reg["Registry"])
                        except KeyError:
                            reglist.append(reg["Schema"])
                        break
                    else:
                        continue

            if not reglist:
                # gen 9 redfish
                regval = [reg.get(arg, None) for arg in ["Registry", "Schema", "Id"]]
                regval = next(
                    (val for val in regval if val and "biosattributeregistry" not in val),
                    None,
                )
                if not regval and reg:
                    reg = reg["@odata.id"].split("/")
                    reg = reg[len(reg) - 2]
                    if "biosattributeregistry" not in reg.lower():
                        reglist.append(reg)
                elif regval:
                    reglist.append(regval)

            for reg in reglist:
                reg = reg.replace("%23", "#")
                messages = self.validation_mgr.get_registry_model(
                    getmsg=True, currtype=reg, searchtype=self.msg_reg_type
                )
                if messages:
                    errmessages.update(messages.get(next(iter(messages)))[regtype.split(".")[-1]])
            if not reglist or not errmessages:
                raise Exception
        except Exception:
            raise EmptyRaiseForEAFP("Unable to find registry schema with provided registry " "type: %s" % regtype)
        else:
            return errmessages



 # Copyright 2020 Hewlett Packard Enterprise Development LP
 #
 # Licensed under the Apache License, Version 2.0 (the "License"); you may
 # not use this file except in compliance with the License. You may obtain
 # a copy of the License at
 #
 #      http://www.apache.org/licenses/LICENSE-2.0
 #
 # Unless required by applicable law or agreed to in writing, software
 # distributed under the License is distributed on an "AS IS" BASIS, WITHOUT
 # WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. See the
 # License for the specific language governing permissions and limitations
 # under the License.

# -*- coding: utf-8 -*-
"""
An example of uploading firmware to the iLO Repository for flashing
"""

import os
import sys
import json
import argparse
from random import randint

from redfish import RedfishClient
from redfish.rest.v1 import ServerDownOrUnreachableError
import logging
from redfish import redfish_logger

from ilorest_util import get_resource_directory

LOGGERFILE = "RedfishApiExamples.log"
LOGGERFORMAT = "%(asctime)s - %(name)s - %(levelname)s - %(message)s"
LOGGER = redfish_logger(LOGGERFILE, LOGGERFORMAT, logging.DEBUG)

def upload_firmware(_redfishobj, firmware_loc, comp_type, update_repo=True, update_target=False):
    resource_instances = get_resource_directory(_redfishobj)

    if DISABLE_RESOURCE_DIR or not resource_instances:
        #resource directory is not available so we will navigate through paths manually
        update_service_uri = _redfishobj.root.obj['UpdateService']['@odata.id']
    else:
        #obtain all account instances from resource directory
        for instance in resource_instances:
            if '#UpdateService.' in instance['@odata.type']:
                update_service_uri = instance['@odata.id']

    print (update_service_uri)

    update_service_response = _redfishobj.get(update_service_uri)

    path = update_service_response.obj.HttpPushUri
    print (path)

    body = []

    json_data = {'UpdateRepository': update_repo, 'UpdateTarget': update_target, 'ETag': 'atag', 'Section': 0}
    session_key = _redfishobj.session_key

    #ImageLocation = "c:\\test"
    #filename = "I41_2.40_08_10_2020.fwpkg"
    #ImagePath = os.path.join(ImageLocation, filename)
    #file_name = os.path.basename(firmware_loc)
    #print (ImagePath)
    with open(firmware_loc, 'rb') as fle:
        output = fle.read()

    session_tuple = ('sessionKey', session_key)
    parameters_tuple = ('parameters', json.dumps(json_data))
    file_tuple = ('file', (filename, output, 'application/octet-stream'))

    #Build the payload from each multipart-form data tuple
    body.append(session_tuple)
    body.append(parameters_tuple)
    body.append(file_tuple)
    #print (body)

    #Create our header dictionary
    header = {'Cookie': 'sessionKey=' + session_key}

    print ('Begin upload....please wait...')
    # We pass the whole list payload to post
    resp = _redfishobj.post(path, body, headers=header)

    if resp.status == 400:
        sys.stderr.write("Failed to upload firmware...")
    elif not resp.status in [200, 201]:
        sys.stderr.write("An http response of '%s' was returned.\n" % resp.status)
    else:
        print("Upload complete!\n")


def create_task(_redfishobj, firmware_loc, tpm_flag=True):

    session_key = _redfishobj.session_key
    #Create our header dictionary
    header = {'Cookie': 'sessionKey=' + session_key}

    updatable_by = ['Uefi']
    task_path = '/redfish/v1/UpdateService/UpdateTaskQueue/'
    file_name = os.path.basename(firmware_loc)

    task_resp = _redfishobj.get(task_path)
    if not task_resp.obj["Members@odata.count"]:
        print ("No current tasks, proceed to create new task")


    update_task = {'Name': 'Update-%s-%s' % (str(randint(0, 1000000)), \
                        file_name), 'Command': 'ApplyUpdate',\
                      'Filename': file_name, 'UpdatableBy': updatable_by, 'TPMOverride': tpm_flag}

    resp = _redfishobj.post(task_path, update_task, headers=header)
    #print (resp)
    if resp.status == 400:
        sys.stderr.write("Failed to create update task queue...")
    elif not resp.status in [200, 201]:
        sys.stderr.write("An http response of '%s' was returned.\n" % resp.status)
    else:
        print("Task Queue Creation complete!\n")


if __name__ == "__main__":

    # Initialize parser
    parser = argparse.ArgumentParser(description = "Script to upload and flash NVMe FW")

    parser.add_argument(
        '-c',
        '--component',
        dest='comp_path',
        action="store",
        required=True,
        help="The path to the firmware file to upload",
        default=None)
    parser.add_argument(
        '-s',
        '--session_key',
        dest='session_key',
        action="store",
        required=False,
        help="Http session key for the server",
        default=None)
    parser.add_argument(
        '-i',
        '--ilo',
        dest='ilo_ip',
        action="store",
        required=False,
        help="iLO IP of the server",
        default=None)
    parser.add_argument(
        '-u',
        '--user',
        dest='ilo_user',
        action="store",
        required=False,
        help="iLO username to login",
        default=None)
    parser.add_argument(
        '-p',
        '--password',
        dest='ilo_pass',
        action="store",
        required=False,
        help="iLO password to log in.",
        default=None)

    options = parser.parse_args()

    system_url = "https://" + options.ilo_ip
    print (system_url)

    # Upload the firmware file to the iLO Repository
    update_repo_flag = True
    # Update the system with the firmware file
    update_target_flag = False

    comp_type = 'C'
    tpm_flag = True

    # flag to force disable resource directory. Resource directory and associated operations are
    # intended for HPE servers.
    DISABLE_RESOURCE_DIR = True
    #system_url = "https://" + "15.146.46.49"
    #session_id = "515b196969f7d879886ee4a2da4ccba8"

    try:
        # Create a Redfish client object
        if options.session_key:
            redfish_obj = RedfishClient(base_url=system_url, session_key= options.session_key)
        else:
            redfish_obj = RedfishClient(base_url=system_url, username=options.ilo_user, password=options.ilo_pass)
        # Login with the Redfish client
        redfish_obj.login()
    except ServerDownOrUnreachableError as excp:
        sys.stderr.write("ERROR: server not reachable or does not support RedFish.\n")
        sys.exit()

    upload_firmware(redfish_obj, options.comp_path, comp_type, update_repo_flag, update_target_flag)
    if comp_type == 'C':
        create_task(redfish_obj, options.comp_path, tpm_flag)

    redfish_obj.logout()



# Copyright 2016-2022 Hewlett Packard Enterprise Development LP
#
# Licensed under the Apache License, Version 2.0 (the "License"); you may
# not use this file except in compliance with the License. You may obtain
# a copy of the License at
#
#      http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS, WITHOUT
# WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. See the
# License for the specific language governing permissions and limitations
# under the License.

# -*- coding: utf-8 -*-
"""
An example of mounting virtual media for HPE iLO systems
"""

import sys
import json
from redfish import RedfishClient
from redfish.rest.v1 import ServerDownOrUnreachableError

from get_resource_directory import get_resource_directory

def mount_virtual_media_iso(_redfishobj, media_type,):

    virtual_media_uri = None
    virtual_media_response = []

    resource_instances = get_resource_directory(_redfishobj)
    if DISABLE_RESOURCE_DIR or not resource_instances:
        #if we do not have a resource directory or want to force it's non use to find the
        #relevant URI
        managers_uri = _redfishobj.root.obj['Managers']['@odata.id']
        managers_response = _redfishobj.get(managers_uri)
        managers_members_uri = next(iter(managers_response.obj['Members']))['@odata.id']
        managers_members_response = _redfishobj.get(managers_members_uri)
        virtual_media_uri = managers_members_response.obj['VirtualMedia']['@odata.id']
    else:
        for instance in resource_instances:
            #Use Resource directory to find the relevant URI
            if '#VirtualMediaCollection.' in instance['@odata.type']:
                virtual_media_uri = instance['@odata.id']

    if virtual_media_uri:
        virtual_media_response = _redfishobj.get(virtual_media_uri)
        for virtual_media_slot in virtual_media_response.obj['Members']:
            data = _redfishobj.get(virtual_media_slot['@odata.id'])
            if media_type in data.dict['MediaTypes']:
                virtual_media_mount_uri = data.obj['Actions']['#VirtualMedia.EjectMedia']['target']
                post_body = {"Action": "HpeiLOVirtualMedia.EjectVirtualMedia"}
                resp = _redfishobj.post(virtual_media_mount_uri, post_body)
                if resp.status == 400:
                    try:
                        print(json.dumps(resp.obj['error']['@Message.ExtendedInfo'], indent=4, \
                                         sort_keys=True))
                    except Exception as excp:
                        sys.stderr.write("A response error occurred, unable to access iLO"
                                         "Extended Message Info...")
                elif resp.status != 200:
                    sys.stderr.write("An http response of \'%s\' was returned.\n" % resp.status)
                else:
                    print("Success!\n")
                    print(json.dumps(resp.dict, indent=4, sort_keys=True))
                break

if __name__ == "__main__":
    # When running on the server locally use the following commented values
    #SYSTEM_URL = None
    #LOGIN_ACCOUNT = None
    #LOGIN_PASSWORD = None

    # When running remotely connect using the secured (https://) address,
    # account name, and password to send https requests
    # SYSTEM_URL acceptable examples:
    # "https://10.0.0.100"
    # "https://ilo.hostname"
    SYSTEM_URL = "https://10.0.0.100"
    LOGIN_ACCOUNT = "admin"
    LOGIN_PASSWORD = "password"

    #specify the type of content the media represents
    MEDIA_TYPE = "CD" #current possible options: Floppy, USBStick, CD, DVD

    # flag to force disable resource directory. Resource directory and associated operations are
    # intended for HPE servers.
    DISABLE_RESOURCE_DIR = False

    try:
        # Create a Redfish client object
        REDFISHOBJ = RedfishClient(base_url=SYSTEM_URL, username=LOGIN_ACCOUNT, \
                                   password=LOGIN_PASSWORD)

        # Login with the Redfish client
        REDFISHOBJ.login()
    except ServerDownOrUnreachableError as excp:
        sys.stderr.write("ERROR: server not reachable or does not support RedFish.\n")
        sys.exit()

    mount_virtual_media_iso(REDFISHOBJ, MEDIA_TYPE)
    REDFISHOBJ.logout()



 # Copyright 2020 Hewlett Packard Enterprise Development LP
 #
 # Licensed under the Apache License, Version 2.0 (the "License"); you may
 # not use this file except in compliance with the License. You may obtain
 # a copy of the License at
 #
 #      http://www.apache.org/licenses/LICENSE-2.0
 #
 # Unless required by applicable law or agreed to in writing, software
 # distributed under the License is distributed on an "AS IS" BASIS, WITHOUT
 # WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. See the
 # License for the specific language governing permissions and limitations
 # under the License.

# -*- coding: utf-8 -*-
"""
An example of getting the ilo information like ilo generation, version and resource directory for HPE iLO systems
"""

import sys
from redfish import RedfishClient
from redfish.rest.v1 import ServerDownOrUnreachableError

def get_resource_directory(redfishobj):

    try:
        resource_uri = redfishobj.root.obj.Oem.Hpe.Links.ResourceDirectory['@odata.id']
    except KeyError:
        sys.stderr.write("Resource directory is only available on HPE servers.\n")
        return None

    response = redfishobj.get(resource_uri)
    resources = []

    if response.status == 200:
        sys.stdout.write("\tFound resource directory at /redfish/v1/resourcedirectory" + "\n\n")
        resources = response.dict["Instances"]
    else:
        sys.stderr.write("\tResource directory missing at /redfish/v1/resourcedirectory" + "\n")

    return resources

def get_gen(_redfishobj):
	rootresp = _redfishobj.root.obj
	#Default iLO 5
	ilogen = 5
	gencompany = next(iter(rootresp.get("Oem", {}).keys()), None) in ('Hpe', 'Hp')
	comp = 'Hp' if gencompany else None
	comp = 'Hpe' if rootresp.get("Oem", {}).get('Hpe', None) else comp
	if comp and next(iter(rootresp.get("Oem", {}).get(comp, {}).get("Manager", {}))).\
																get('ManagerType', None):
		ilogen = next(iter(rootresp.get("Oem", {}).get(comp, {}).get("Manager", {})))\
																		.get("ManagerType")
		ilover = next(iter(rootresp.get("Oem", {}).get(comp, {}).get("Manager", {}))).\
															  get("ManagerFirmwareVersion")
		if ilogen.split(' ')[-1] == "CM":
			# Assume iLO 4 types in Moonshot
			ilogen = 4
			iloversion = None
		else:
			ilogen = ilogen.split(' ')[1]
			iloversion = float(ilogen.split(' ')[-1] + '.' + \
											''.join(ilover.split('.')))
	return (ilogen, iloversion)



import sys
import json
from redfish import RedfishClient
from redfish.rest.v1 import ServerDownOrUnreachableError


def configure_snmp(_redfishobj):

    snmp_service_uri = None
    managers_uri = _redfishobj.root.obj['Managers']['@odata.id']
    managers_response = _redfishobj.get(managers_uri)
    managers_members_uri = next(iter(managers_response.obj['Members']))['@odata.id']
    managers_members_response = _redfishobj.get(managers_members_uri)
    # print(managers_members_response.obj['Oem']['Hpe']['Links'])
    # exit(0)
    snmp_service_uri = managers_members_response.obj.Oem.Hpe.Links['SNMPService']['@odata.id']

    if snmp_service_uri:
        # resp = _redfishobj.get(snmp_service_uri)
        # print(json.dumps(resp.dict, indent=4, sort_keys=True))
        # exit(0)
        # TrapSourceHostname : Manager or System
        body = {"TrapSourceHostname": "System"}
        resp = _redfishobj.patch(snmp_service_uri, body)

        if resp.status == 400:
            try:
                print(json.dumps(resp.obj['error']['@Message.ExtendedInfo'], indent=4,
                                 sort_keys=True))
            except Exception as excp:
                sys.stderr.write("A response error occurred, unable to access iLO Extended "
                                 "Message Info...")
        elif resp.status != 200:
            sys.stderr.write("An http response of \'%s\' was returned.\n" % resp.status)
        else:
            print("Success!\n")
            # print(json.dumps(resp.dict, indent=4, sort_keys=True))


if __name__ == "__main__":

    if len(sys.argv) <= 3:
        sys.stderr.write("ERROR: We need 3 arguments. \nUsage: change-snmp.py "
                         "ILOIPADDR ILOUSER ILOPASSWORD \n")
        sys.exit()

    SYSTEM_URL = "https://" + sys.argv[1]
    LOGIN_ACCOUNT = sys.argv[2]
    LOGIN_PASSWORD = sys.argv[3]

    try:
        REDFISHOBJ = RedfishClient(base_url=SYSTEM_URL, username=LOGIN_ACCOUNT,
                                   password=LOGIN_PASSWORD)
        REDFISHOBJ.login()
    except ServerDownOrUnreachableError as excp:
        sys.stderr.write("ERROR: server not reachable or does not support RedFish.\n")
        sys.exit()

    configure_snmp(REDFISHOBJ)
    REDFISHOBJ.logout()


 # Copyright 2020 Hewlett Packard Enterprise Development LP
 #
 # Licensed under the Apache License, Version 2.0 (the "License"); you may
 # not use this file except in compliance with the License. You may obtain
 # a copy of the License at
 #
 #      http://www.apache.org/licenses/LICENSE-2.0
 #
 # Unless required by applicable law or agreed to in writing, software
 # distributed under the License is distributed on an "AS IS" BASIS, WITHOUT
 # WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. See the
 # License for the specific language governing permissions and limitations
 # under the License.

# -*- coding: utf-8 -*-
"""
An example of clearing IEL or IML Logs for HPE iLO systems
"""

import sys
import json
from redfish import RedfishClient
from redfish.rest.v1 import ServerDownOrUnreachableError

from get_resource_directory import get_resource_directory

def clear_ilo_event_log(_redfishobj, clear_IML_IEL):

    clear_log_services_uri = []

    resource_instances = get_resource_directory(_redfishobj)
    if DISABLE_RESOURCE_DIR or not resource_instances:
        #if we do not have a resource directory or want to force it's non use to find the
        #relevant URI
        systems_uri = _redfishobj.root.obj['Systems']['@odata.id']
        systems_response = _redfishobj.get(systems_uri)
        systems_members_uri = next(iter(systems_response.obj['Members']))['@odata.id']
        systems_members_response = _redfishobj.get(systems_members_uri)
        log_services_uri = systems_members_response.obj['LogServices']['@odata.id']
        log_services_response = _redfishobj.get(log_services_uri)
        log_services_uris = log_services_response.obj['Members']
        for log_services_uri in log_services_uris:
            log_services_response = _redfishobj.get(log_services_uri['@odata.id'])
            clear_log_services_uri.append(log_services_response.obj['Actions']\
                                                                ['#LogService.ClearLog']['target'])
    else:
        for instance in resource_instances:
            #Use Resource directory to find the relevant URI
            if '#LogService.' in instance['@odata.type']:
                log_service_uri = instance['@odata.id']
                clear_log_services_uri.append(_redfishobj.get(log_service_uri).dict['Actions']\
                                                                ['#LogService.ClearLog']['target'])

    if clear_log_services_uri:
        body = {"Action": "LogService.ClearLog"}
        for path in clear_log_services_uri:
            if ("IEL" in clear_IML_IEL and "IEL" in path) or ("IML" in clear_IML_IEL and \
                                                                                    "IML" in path):
                if "IEL" in path:
                    sys.stdout.write("Clearing IEL log.\n")
                else:
                    sys.stdout.write("Clearing IML log.\n")
                resp = _redfishobj.post(path, body)
            else:
                continue
            #If iLO responds with soemthing outside of 200 or 201 then lets check the iLO extended
            #info error message to see what went wrong
            if resp.status == 400:
                try:
                    print(json.dumps(resp.obj['error']['@Message.ExtendedInfo'], indent=4, \
                                                                                    sort_keys=True))
                except Exception as excp:
                    sys.stderr.write("A response error occurred, unable to access iLO "\
                                     "Extended Message Info...\n")
            elif resp.status != 200:
                sys.stderr.write("An http response of \'%s\' was returned.\n" % resp.status)
            else:
                print("Success!\n")
                print(json.dumps(resp.dict, indent=4, sort_keys=True))

if __name__ == "__main__":
    # When running on the server locally use the following commented values
    #SYSTEM_URL = None
    #LOGIN_ACCOUNT = None
    #LOGIN_PASSWORD = None

    # When running remotely connect using the secured (https://) address,
    # account name, and password to send https requests
    # SYSTEM_URL acceptable examples:
    # "https://10.0.0.100"
    # "https://ilo.hostname"
    SYSTEM_URL = "https://10.0.0.100"
    LOGIN_ACCOUNT = "admin"
    LOGIN_PASSWORD = "password"

    # flag to force disable resource directory. Resource directory and associated operations are
    # intended for HPE servers.
    DISABLE_RESOURCE_DIR = False
    CLEAR_IML_IEL = "IML" # provide either the string IML or IEL

    try:
        # Create a Redfish client object
        REDFISHOBJ = RedfishClient(base_url=SYSTEM_URL, username=LOGIN_ACCOUNT, \
                                                                            password=LOGIN_PASSWORD)
        # Login with the Redfish client
        REDFISHOBJ.login()
    except ServerDownOrUnreachableError as excp:
        sys.stderr.write("ERROR: server not reachable or does not support RedFish.\n")
        sys.exit()

    clear_ilo_event_log(REDFISHOBJ, CLEAR_IML_IEL)
    REDFISHOBJ.logout()



###
# Copyright 2020 Hewlett Packard Enterprise, Inc. All rights reserved.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#  http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
###

# -*- coding: utf-8 -*-
"""Handles schema and registry gathering as well as schema parsing and validation."""

# ---------Imports---------

import json
import logging
import re
import textwrap

import six

from redfish.rest.containers import RisObject

from .sharedtypes import JSONEncoder

# ---------End of imports---------


# ---------Debug logger---------

LOGGER = logging.getLogger(__name__)


# ---------End of debug logger---------


class InvalidPathsError(Exception):
    """Raised when requested path is not found"""

    pass


class RegistryValidationError(Exception):
    """Registration Validation Class Error"""

    def __init__(self, msg, regentry=None, selector=None):
        super(RegistryValidationError, self).__init__(msg)
        self.reg = regentry
        self.sel = selector
        self.message = msg


class UnknownValidatorError(Exception):
    """Raised when we find an attribute type that we don't know how to validate."""

    pass


class ValidationManager(object):
    """Keeps track of all the schemas and registries and provides helpers
    to simplify validation."""

    def __init__(self, monolith, defines=None):
        """init of ValidationManager"""
        super(ValidationManager, self).__init__()

        self._schemaid = Typepathforval.typepath.schemapath
        self._regid = Typepathforval.typepath.regpath

        self._classes = list()
        self._classpaths = list()
        # type and path defines object
        self.defines = defines
        self.monolith = monolith
        # error
        self._errors = list()
        self._warnings = list()
        self.updatevalidationdata()

    @property
    def errors(self):
        """All errors found by the last validation."""
        return self._errors

    @property
    def warnings(self):
        """All warnings found by the last validation."""
        return self._warnings

    def reset_errors_warnings(self):
        """Resets warnings and errors, getting ready for the next validation."""
        self._errors = list()
        self._warnings = list()

    def updatevalidationdata(self):
        """Loads the types into the validation manager from monolith."""
        monolith = self.monolith
        for instance in monolith.iter():
            if (
                (
                    x.lower() in instance.maj_type.lower()
                    for x in (
                        self.defines.defs.schemafilecollectiontype,
                        "Collection.",
                        self.defines.defs.regfilecollectiontype,
                    )
                )
                and any(x.lower() in instance.path.lower() for x in (self._schemaid, self._regid))
                and instance
                and instance.path not in self._classpaths
            ):
                self._classpaths.append(instance.path)
                self._classes.append(instance.resp.dict)

    def find_prop(self, propname, latestschema=False, proppath=None):
        """Searches through all locations and returns the schema
        found for the provided propname type.

        :param propname: String containing the schema name.
        :type propname: str
        :param proppath: String containing the schema path if you wish to use that instead.
        :type proppath: str
        :param latestschema: Flag to determine if we should drop the schema version when we try to
                             match schema information. If True, the version will be dropped.
        :type latestschema: bool
        """
        if proppath:
            self.monolith.load(path=proppath, crawl=False, loadtype="ref")
            return True
        for cls in self._classes:
            found = self.find_property(propname, cls=cls, latestschema=latestschema)
            if found:
                return found
        return None

    def itermems(self, membername=None):
        """Searches through all locations and yields each entry.

        :param membername: string containing the registry name. If not passed we use the typedefines
                           string by default.
        :type membername: str
        """
        if not membername:
            membername = self.defines.defs.collectionstring
        for items in self._classes:
            for item in items[membername]:
                yield item

    def iterregmems(self, membername=None):
        """Searches through all registries and yields each entry.

        :param membername: string containing the registry name. If not passed we use the typedefines
                           string by default.
        :type membername: str
        """
        if not membername:
            membername = self.defines.defs.collectionstring
        for items in self._classes:
            if "registr" in items["Name"].lower():  # For Gen9 type/name issue
                for item in items[membername]:
                    yield item

    def iterschemamems(self, membername=None):
        """Searches through all schemas and yields each entry

        :param membername: string containing the registry name. If not passed we use the typedefines
                           string by default.
        :type membername: str
        """
        if not membername:
            membername = self.defines.defs.collectionstring
        for items in self._classes:
            if "schema" in items["Name"].lower():  # For Gen9 type/name issue
                for item in items[membername]:
                    yield item

    def find_property(self, propname, cls=None, latestschema=False):
        """Returns iLO/BIOS registries/schemas

        :param propname: string containing the registry/schema name.
        :type propname: str
        :param cls: self._classes list of dictionaries.
        :type cls: list
        :param latestschema: flag to drop the versioning in the type string.
        :type latestschema: bool.

        :returns: iLO/BIOS registries/schemas that match the supplied name.
        """
        result = []
        dataloc = cls.get("Items", None)
        dataloc = cls.get("Members", None) if not dataloc else dataloc
        keyword = "Schema"
        if dataloc and isinstance(dataloc, list):
            _ = propname.split(".")[0].strip("#")
            propname = propname.split(".")[0].strip("#") if latestschema else propname
            for entry in dataloc:
                if entry:
                    if "Schema" in entry:
                        if propname.lower() in entry["Schema"].lower():
                            result.append(entry)
                    elif "Registry" in entry:
                        if (
                            propname.lower() in entry["@odata.id"].lower()
                            and propname.lower() in entry["Registry"].lower()
                        ):
                            result.append(entry)
                            keyword = "Registry"
                    else:
                        if "@odata.id" in entry:
                            reglink = entry["@odata.id"].split("/")
                            reglink = reglink[len(reglink) - 2]
                            if reglink.lower().startswith(propname.lower()):
                                self.monolith.load(path=entry["@odata.id"], crawl=False)
                                result.append(self.monolith.paths[entry["@odata.id"]].dict)

        if result:
            result = max(
                result,
                key=lambda res: res[Typepathforval.typepath.defs.hrefstring]
                if res.get(Typepathforval.typepath.defs.hrefstring, None)
                else res[keyword],
            )
            schemapath = self.geturidict(result["Location"][0])
            self.monolith.load(path=schemapath, crawl=False, loadtype="ref")
            return result

    def geturidict(self, locationobj):
        """Return the external reference link.

        :param locationobj: Dictionary to get the URI reference from.
        :type locationobj: dict
        """
        if Typepathforval.typepath.defs.isgen10:
            try:
                return locationobj["Uri"]
            except KeyError:
                raise InvalidPathsError("Error accessing Uri path!/n")
        elif Typepathforval.typepath.defs.isgen9:
            try:
                return locationobj["Uri"]["extref"]
            except KeyError:
                raise InvalidPathsError("Error accessing extref path!/n")

    def validatedict(
        self,
        tdict,
        currtype=None,
        proppath=None,
        latestschema=False,
        searchtype=None,
        monolith=None,
        reg=None,
        unique=None,
    ):
        """Load the schema file and validate tdict against it.

        :param tdict: the dictionary to test against.
        :type tdict: dict
        :param currtype: String containing the type the tdict dictionary is.
        :type currtype: str
        :param proppath: String containing the schema path of the tdict dictionary if you wish to
                         use that instead.
        :type proppath: str
        :param latestschema: Flag to determine if we should drop the schema version when we try to
                             match schema information. If True, the version will be dropped.
        :type latestschema: bool
        :param searchtype: Include the attribute registry of you are validating a bios registry.
        :type searchtype: str
        :param monolith: Full data model retrieved from server.
        :type monolith: dict
        :param unique: Flag to override for skipping unique properties.
        :type unique: bool
        :param reg: Registry entry of the given attribute. If this is not provided we will attempt
                    to search based on the searchtype and currtype/proppath arguments.
        :type reg: dict.
        :returns: returns an error list.
        """
        if not reg:
            reg = self.get_registry_model(
                currtype=currtype,
                searchtype=searchtype,
                proppath=proppath,
                latestschema=latestschema,
            )

        if reg:
            list(
                map(
                    lambda x: self.checkreadunique(
                        tdict,
                        x,
                        reg=reg,
                        warnings=self._warnings,
                        unique=unique,
                        searchtype=searchtype,
                    ),
                    list(tdict.keys()),
                )
            )
            orireg = reg.copy()
            ttdict = {key: val for key, val in list(tdict.items()) if not isinstance(val, (dict, list))}
            results = reg.validate_attribute_values(ttdict)
            self._errors.extend(results)

            for ki, val in list(tdict.items()):
                if ki in ttdict:
                    tdict[ki] = ttdict[ki]
                    continue
                reg = orireg.copy()
                valexists = False
                if val and isinstance(val, list):
                    valexists = True
                    # TODO: only validates if its a single dict within list
                    if len(val) == 1 and isinstance(val[0], dict):
                        treg = self.nestedreg(reg=reg, args=[ki])
                        self.validatedict(
                            val[0],
                            unique=unique,
                            monolith=monolith,
                            reg=treg,
                            currtype=currtype,
                            searchtype=searchtype,
                        )
                    else:
                        continue
                elif val and isinstance(val, dict):
                    valexists = True
                    treg = self.nestedreg(reg=reg, args=[ki])
                    self.validatedict(
                        val,
                        monolith=monolith,
                        reg=treg,
                        unique=unique,
                        searchtype=searchtype,
                    )
                if not val and valexists:
                    del tdict[ki]

        else:
            self._errors.append(RegistryValidationError("Unable to locate registry model"))
        return self._errors, self._warnings

    def checkreadunique(self, tdict, tkey, reg=None, warnings=None, unique=None, searchtype=None):
        """Check for and remove the readonly and unique attributes if required.

        :param tdict: the dictionary to test against.
        :type tdict: dict.
        :param tkey: The attribute key value to be tested.
        :type tkey: str.
        :param warnings: list containing found warnings.
        :type warnings: list.
        :param unique: flag to determine override for unique properties.
        :type unique: str.
        :param reg: Registry entry of the given attribute.
        :type reg: dict.
        :returns: returns boolean.
        """
        if "Attributes" in tdict:
            return False
            # key = list(tdict[tkey])[0]
            # reg = reg["Attributes"][key]
        else:
            if tkey in reg:
                reg = reg[tkey]

        if not unique and reg.get("IsSystemUniqueProperty", None):
            if tdict[tkey] and not isinstance(tdict[tkey], bool):
                self._warnings.append("Property '%s' is unique and override not authorized. Skipping...\n" % str(tkey))
                del tdict[tkey]
                return True
        if not reg.get("ReadOnly") or (reg.get(tkey, None) and not reg[tkey].get("readonly")):
            if unique and reg.get("IsSystemUniqueProperty", None):
                self._warnings.append("Property '%s' is unique, but override authorized...Patching..\n" % str(tkey))
                return False
        # if not searchtype or (reg.get("ReadOnly") or (reg.get(tkey)
        #                                                and reg[tkey].get("readonly"))):
        #    self._warnings.append("Property is read-only. skipping... '%s'" % str(tkey))
        #    del tdict[tkey]
        #    return True
        elif reg.get("ReadOnly") or (reg.get(tkey) and reg[tkey].get("readonly")):
            if tdict[tkey]:
                self._warnings.append("Property '%s' is read-only. Skipping...\n" % str(tkey))
                del tdict[tkey]
                return True
        else:
            return False

    def get_registry_model(
        self,
        currtype=None,
        proppath=None,
        getmsg=False,
        searchtype=None,
        newarg=None,
        latestschema=False,
    ):
        """Loads the schema file and find the registry model if available. A registry model is a
        object built for schema/bios registry data.

        :param currtype: Type selection string.
        :type currtype: dict.
        :param proppath: String containing the schema path if you wish to use that instead.
        :type proppath: str
        :param getmsg: Flag to determine if commit should be skipped.
        :type getmsg: bool
        :param searchtype: Include the attribute registry of you are validating a bios registry.
        :type searchtype: str
        :param newarg: List of multi level properties to be modified.
        :type newarg: list
        :param latestschema: Flag to determine if we should drop the schema version when we try to
                             match schema information. If True, the version will be dropped.
        :type latestschema: bool
        :returns: Schema in object form called a registry object.
        """
        regdict = None
        monolith = self.monolith
        currtype = currtype.split("#")[-1].split(".")[0] + "." if currtype and latestschema else currtype
        if (
            not currtype
            or not self.find_prop(
                currtype,
                latestschema=latestschema,
                proppath=proppath if not searchtype else None,
            )
        ) and (not searchtype):
            self._errors.append(RegistryValidationError("Location info is missing.\n"))
            return None
        if not searchtype:
            searchtype = "object"

        try:
            for instance in monolith.iter(searchtype):
                if (
                    (searchtype == Typepathforval.typepath.defs.attributeregtype)
                    or (
                        searchtype == "object"
                        and any(
                            currtype in xtitle
                            for xtitle in (
                                instance.resp.dict.get("title", ""),
                                instance.resp.dict.get("oldtitle", ""),
                            )
                        )
                    )
                    or (
                        searchtype != "object"
                        and currtype.split("#")[-1].split(".")[0] == instance.dict.get("RegistryPrefix", "")
                    )
                ):
                    regdict = instance.resp.dict
                    break
        except BaseException:
            pass

        if not regdict:
            self._errors.append(RegistryValidationError("Location data is empty.\n"))
            return None

        jsonreg = json.loads(json.dumps(regdict, indent=2, cls=JSONEncoder))

        if getmsg:
            return {jsonreg["RegistryPrefix"]: jsonreg["Messages"]}

        # This was done for bios registry model compatibility
        if "RegistryEntries" in jsonreg:
            regitem = jsonreg["RegistryEntries"]
            if "Attributes" in regitem:
                newitem = {item[Typepathforval.typepath.defs.attributenametype]: item for item in regitem["Attributes"]}
                regitem["Attributes"] = newitem
                if not Typepathforval.typepath.flagiften:
                    del regitem["Attributes"]
                    newitem.update(regitem)
                    regitem = newitem
                reg = HpPropertiesRegistry.parse(regitem)
            return self.nestedreg(reg=reg, args=newarg) if newarg else reg

        if "properties" in jsonreg:
            regitem = jsonreg["properties"]
            if "Properties" in regitem:
                regitem.update(regitem["Properties"])
                del regitem["Properties"]
            reg = HpPropertiesRegistry.parse(regitem)

            return self.nestedreg(reg=reg, args=newarg) if newarg else reg

    def nestedreg(self, reg=None, args=None):
        """Go through the registry entry to find the required nested attribute.

        :param reg: Registry entry of the given attribute.
        :type reg: dict
        :param args: List of multi level properties to be modified.
        :type args: list
        :returns: dict of Registry entry
        """
        for arg in args:
            try:
                arg = next((key for key in list(reg.keys()) if key.lower() == arg.lower()), None)
                if not arg:
                    return None
                if ("properties" in reg[arg].keys()) and ("patternProperties" in reg[arg].keys()):
                    reg[arg]["properties"].update(reg[arg]["patternProperties"])
                    reg = reg[arg]["properties"]
                elif "oneOf" in reg[arg]:
                    oneof = reg[arg]["oneOf"]
                    for item in oneof:
                        reg = item["properties"]
                elif (
                    "type" in reg[arg]
                    and reg[arg]["type"] == "array"
                    and "items" in reg[arg]
                    and "properties" in reg[arg]["items"]
                ):
                    reg = reg[arg]["items"]["properties"]
                elif ("properties" not in reg[arg].keys()) or ("patternProperties" in reg[arg].keys()):
                    reg = reg[arg]
                else:
                    reg = reg[arg]["properties"]
            except:
                try:
                    reg = reg[arg]["patternProperties"]
                except:
                    return None
        return reg


class HpPropertiesRegistry(RisObject):
    """Models a schema or bios attribute registry. Registry model."""

    def __init__(self, d):
        super(HpPropertiesRegistry, self).__init__(d)

    def validate_attribute_values(self, tdict):
        """Look for tdict in the attribute list and attempt to validate its value.

        :param tdict: the dictionary to test against.
        :type tdict: dict
        :returns: A validated list
        """
        result = list()

        for tkey in tdict:
            if tkey not in self:
                # Added for Gen 9 Bios properties not in registry
                continue
            elif self[tkey] and (checkattr(self[tkey], "type") or checkattr(self[tkey], "Type")):
                keyval = list()
                keyval.append(tdict[tkey])
                temp = self.validate_attribute(self[tkey], keyval, tkey)
                tdict[tkey] = keyval[0]

                for err in temp:
                    if isinstance(err, RegistryValidationError):
                        if err.reg:
                            err.sel = tkey

                result.extend(temp)

        return result

    def get_validator(self, attrname, newargs=None, oneof=None):
        """Returns attribute validator type.

        :param attrname: attribute name to validate. Ex: In A/B/C, this will be A.
        :type attrname: str
        :param newargs: List of multi level properties to be modified. Ex: In A/B/C this will be
                        a list of B and C.
        :type newargs: list
        :param oneof: Special string for "oneof" options within validation.
        :type oneof: string
        :returns: The validator type class for the property passed.
        """
        if oneof:
            self = oneof

        if newargs:
            for arg in newargs:
                try:
                    self = self["properties"]
                except Exception:
                    pass

                if not checkattr(self, arg):
                    return None
                elif not arg == newargs[-1]:
                    self = self[arg]

        if not checkattr(self, attrname):
            return None

        validator = None
        if EnumValidator.is_type(self[attrname]):
            validator = EnumValidator.parse(self[attrname])
        elif StringValidator.is_type(self[attrname]):
            validator = StringValidator.parse(self[attrname])
        elif ObjectValidator.is_type(self[attrname]):
            validator = ObjectValidator.parse(self[attrname])
        elif IntegerValidator.is_type(self[attrname]):
            validator = IntegerValidator.parse(self[attrname])
        elif BoolValidator.is_type(self[attrname]):
            validator = BoolValidator.parse(self[attrname])
        elif PasswordValidator.is_type(self[attrname]):
            validator = PasswordValidator.parse(self[attrname])
        elif "oneOf" in list(self[attrname].keys()):
            for item in self[attrname]["oneOf"]:
                validator = self.get_validator(attrname, newargs, HpPropertiesRegistry({attrname: item}))
                if validator:
                    break
        return validator

    def validate_attribute(self, attrentry, attrvallist, name):
        """Function to validate attribute against its schema.

        :param attrentry: Key of property to validate.
        :type attrentry: str
        :param attrval: Value of Key to validate.
        :type attrval: str
        :param name: Clean name for outputting information to users.
        :type name: str
        :returns: returns list with validated attributes
        """
        result = list()
        validator = None
        if self.nulltypevalidationcheck(attrval=attrvallist[0], attrentry=attrentry):
            return result

        if EnumValidator.is_type(attrentry):
            validator = EnumValidator.parse(attrentry)
        elif StringValidator.is_type(attrentry):
            validator = StringValidator.parse(attrentry)
        elif IntegerValidator.is_type(attrentry):
            validator = IntegerValidator.parse(attrentry)
        elif BoolValidator.is_type(attrentry):
            validator = BoolValidator.parse(attrentry)
        elif ObjectValidator.is_type(attrentry):
            validator = ObjectValidator.parse(attrentry)
        elif PasswordValidator.is_type(attrentry):
            validator = PasswordValidator.parse(attrentry)
        else:
            raise UnknownValidatorError(attrentry)

        if validator:
            result.extend(validator.is_array(attrentry, attrvallist, name))
            result.extend(validator.validate(attrvallist, name))
        return result

    def nulltypevalidationcheck(self, attrval=None, attrentry=None):
        """Function to validate null attributes against iLO schema

        :param attrentry: Key of property to validate.
        :type attrentry: str
        :param attrval: Value of Key to validate.
        :type attrval: str
        :returns: True if entry is null and valid.
        """
        if "type" in attrentry and attrval is None:
            if isinstance(attrentry["type"], list):
                for item in attrentry["type"]:
                    if item.lower() == "null":
                        return True
        return False


class BaseValidator(RisObject):
    """Base class for all validators."""

    def __init__(self, d):
        super(BaseValidator, self).__init__(d)

    def validate(self):
        """Overridable function for validation"""
        raise RuntimeError("You must override this method in your derived class")

    def common_print_help(self, name):
        """Common human readable schema data.

        :param name: clean name for outputting.
        :type name: str
        :returns: A human readable string of schema data.
        """
        outdata = ""
        wrapper = textwrap.TextWrapper()
        wrapper.initial_indent = " " * 4
        wrapper.subsequent_indent = " " * 4

        outdata += "\nNAME\n"
        outdata += "%s\n" % wrapper.fill("%s" % name)
        outdata += "\n"

        if "DisplayName" in self:
            outdata += "\nDISPLAY NAME\n"
            outdata += "%s\n" % wrapper.fill("%(DisplayName)s" % self)
            outdata += "\n"

        if "description" in self:
            outdata += "\nDESCRIPTION\n"
            outdata += "%s\n" % wrapper.fill("%(description)s" % self)
            outdata += "\n"

        if "HelpText" in self:
            outdata += "\nHELP TEXT\n"
            outdata += "%s\n" % wrapper.fill("%(HelpText)s" % self)
            outdata += "\n"

        if "WarningText" in self:
            outdata += "\n************************************************\n"
            outdata += "\nWARNING\n"
            outdata += "%s\n" % wrapper.fill("%(WarningText)s" % self)
            outdata += "\n\n**********************************************\n"
            outdata += "\n"

        if "type" in self and isinstance(self["type"], list):
            outdata += "\nTYPE\n"
            for item in self["type"]:
                outdata += "%s\n" % wrapper.fill("%s" % item)
            outdata += "\n"
        elif "type" in self:
            outdata += "\nTYPE\n"
            outdata += "%s\n" % wrapper.fill("%(type)s" % self)
            outdata += "\n"
        elif "Type" in self:
            outdata += "\nTYPE\n"
            outdata += "%s\n" % wrapper.fill("%(Type)s" % self)
            outdata += "\n"

        if "ReadOnly" in self:
            outdata += "\nREAD-ONLY\n"
            outdata += "%s\n" % wrapper.fill("%(ReadOnly)s" % self)
            outdata += "\n"
        elif "readonly" in self:
            outdata += "\nREAD-ONLY\n"
            outdata += "%s\n" % wrapper.fill("%(readonly)s" % self)
            outdata += "\n"
        return outdata

    def is_arrtype(self, attrentry):
        """Validate that the type is an array.

        :param attrentry: Registry model entry used for validation.
        :type attrentry: dict
        :returns: A boolean based on whether type is an array.
        """
        if "type" in attrentry and attrentry["type"] == "array":
            return True
        return False

    def is_array(self, attrentry, arrval, name):
        """Validate that the given value is an array type.

        :param attrentry: Registry model entry used for validation.
        :type attrentry: dict
        :param attrval: Value of Key to validate.
        :type attrval: str
        :returns: A boolean based on whether type is array and the value is valid for array type.
        """
        result = []

        if self.is_arrtype(attrentry):
            if isinstance(
                arrval[0],
                (
                    frozenset,
                    list,
                    set,
                    tuple,
                ),
            ):
                return []
            else:
                result.append(
                    RegistryValidationError(
                        "'%s' is not a valid setting " "for '%s', expecting an array" % (arrval[0], name),
                        regentry=self,
                    )
                )
        return result


class EnumValidator(BaseValidator):
    """Enum validator class"""

    def __init__(self, d):
        super(EnumValidator, self).__init__(d)

    @staticmethod
    def is_type(attrentry):
        """Validate that the type is enumeration.

        :param attrentry: Registry model entry used for validation.
        :type attrentry: dict
        :returns: A boolean based on whether type is eneumeration.
        """
        if "type" in attrentry:
            if isinstance(attrentry["type"], list):
                for item in attrentry["type"]:
                    if item.lower() == "enumeration":
                        return True
                    elif "enum" in attrentry and item.lower() == "string":
                        return True
            elif "enum" in attrentry and attrentry["type"] == "array":
                for key, value in attrentry["items"].items():
                    if key.lower() == "type" and value.lower() == "string":
                        return True
            else:
                if attrentry["type"].lower() == "enumeration":
                    return True
                elif "enum" in attrentry and attrentry["type"].lower() == "string":
                    return True
        elif "Type" in attrentry:
            if attrentry["Type"].lower() == "enumeration":
                return True

        return False

    def validate(self, keyval, name):
        """Validate against schemas.

        :param keyval: New value to be used for validation in a list
        :type keyval: list
        :param name: Clean name for outputting human readable info.
        :type name: str
        :returns: An error if validation fails.
        """
        result = list()
        newval = keyval[0]

        try:
            for possibleval in self.enum:
                if (
                    possibleval
                    and (
                        isinstance(possibleval, type(newval))
                        or (isinstance(possibleval, six.string_types) and isinstance(newval, six.string_types))
                    )
                    and possibleval.lower() == str(newval).lower()
                ):
                    keyval[0] = possibleval
                    return result
        except Exception:
            for possibleval in self.Value:
                if possibleval.ValueName.lower() == str(newval).lower():
                    keyval[0] = possibleval.ValueName
                    return result

        result.append(
            RegistryValidationError("'%s' is not a valid setting " "for '%s'" % (newval, name), regentry=self)
        )

        return result

    def print_help(self, name):
        """Human readable schema information specific to Enum data.

        :param name: Clean name for outputting human readable info.
        :type name: str
        :returns: A human readable string of schema data.
        """
        outdata = self.common_print_help(name)
        outdata += "\nPOSSIBLE VALUES\n"
        try:
            for possibleval in self.enum:
                outdata += "    %s\n" % possibleval
        except Exception:
            for possibleval in self.Value:
                outdata += "    %(ValueName)s\n" % possibleval
        outdata += "\n"
        return outdata


class BoolValidator(BaseValidator):
    """Bool validator class"""

    def __init__(self, d):
        super(BoolValidator, self).__init__(d)

    @staticmethod
    def is_type(attrentry):
        """Validate that the type is boolean.

        :param attrentry: Registry model entry used for validation.
        :type attrentry: dict
        :returns: A boolean based on whether type is boolean.
        """
        if "type" in attrentry:
            if isinstance(attrentry["type"], list):
                for item in attrentry["type"]:
                    if item.lower() == "boolean":
                        return True
            elif attrentry["type"] == "array":
                for key, value in attrentry["items"].items():
                    if key.lower() == "type" and value.lower() == "boolean":
                        return True
            else:
                if attrentry["type"].lower() == "boolean":
                    return True
        elif "Type" in attrentry:
            if attrentry["Type"].lower() == "boolean":
                return True

        return False

    def validate(self, newval, name):
        """Validate against schemas.

        :param newval: New value to be used for validation in a list
        :type newval: list
        :param name: Clean name for outputting human readable info.
        :type name: str
        :returns: An error if value is invalid.
        """
        result = list()
        if newval[0] is False or newval[0] is True:
            return result

        result.append(
            RegistryValidationError("'%s' is not a valid setting for '%s'" % (newval[0], name), regentry=self)
        )

        return result

    def print_help(self, name):
        """Human readable schema information specific to Boolean data.

        :param name: Clean name for outputting human readable info.
        :type name: str
        :returns: A human readable string of schema data.
        """
        outdata = self.common_print_help(name)
        outdata += "\nPOSSIBLE VALUES\n"
        outdata += "    True or False\n"
        outdata += "\n"
        return outdata


class StringValidator(BaseValidator):
    """Constructor"""

    def __init__(self, d):
        super(StringValidator, self).__init__(d)

    @staticmethod
    def is_type(attrentry):
        """Validate that the type is string.

        :param attrentry: Registry model entry used for validation.
        :type attrentry: dict
        :returns: A boolean based on whether type is string.
        """
        if "type" in attrentry:
            if isinstance(attrentry["type"], list):
                for item in attrentry["type"]:
                    if item.lower() == "string":
                        return True
            elif attrentry["type"] == "array":
                for key, value in attrentry["items"].items():
                    if key.lower() == "type" and "string" in value:
                        return True
            else:
                if attrentry["type"].lower() == "string":
                    return True
        elif "Type" in attrentry:
            if attrentry["Type"].lower() == "string":
                return True

        return False

    def validate(self, newvallist, _):
        """Validate against schemas.

        :param newvallist: New value to be used for validation in a list
        :type newvallist: list
        :returns: An error if validation fails criteria.
        """
        newval = newvallist[0]
        result = list()
        namestr = Typepathforval.typepath.defs.attributenametype
        if not isinstance(newval, str):
            result.append(RegistryValidationError("Given value must be a string"))
            return result
        if "MinLength" in self:
            if len(newval) < int(self["MinLength"]):
                result.append(
                    RegistryValidationError(
                        "'%s' must be at least '%s' characters long" % (self[namestr], int(self["MinLength"])),
                        regentry=self,
                    )
                )

        if "MaxLength" in self:
            if len(newval) > int(self["MaxLength"]):
                result.append(
                    RegistryValidationError(
                        "'%s' must be less than '%s' characters long" % (self[namestr], int(self["MaxLength"])),
                        regentry=self,
                    )
                )

        if "ValueExpression" in self:
            if self["ValueExpression"]:
                pat = re.compile(self["ValueExpression"])
                if newval and not pat.match(newval):
                    result.append(
                        RegistryValidationError(
                            "'%s' must match the regular expression " "'%s'" % (self[namestr], self["ValueExpression"]),
                            regentry=self,
                        )
                    )

        return result

    def print_help(self, name):
        """Human readable schema information specific to String data.

        :param name: Clean name for outputting human readable info.
        :type name: str
        :returns: A human readable string of schema data.
        """
        wrapper = textwrap.TextWrapper()
        wrapper.initial_indent = " " * 4
        wrapper.subsequent_indent = " " * 4
        outdata = self.common_print_help(name)
        if "MinLength" in self:
            outdata += "\nMIN LENGTH\n"
            outdata += "%s" % wrapper.fill("%(MinLength)s" % self)
            outdata += "\n"

        if "MaxLength" in self:
            outdata += "\nMAX LENGTH\n"
            outdata += "%s" % wrapper.fill("%(MaxLength)s" % self)
            outdata += "\n"
        return outdata


class IntegerValidator(BaseValidator):
    """Interger validator class"""

    def __init__(self, d):
        super(IntegerValidator, self).__init__(d)

    @staticmethod
    def is_type(attrentry):
        """Validate that the type is integer.

        :param attrentry: Registry model entry used for validation.
        :type attrentry: dict
        :returns: A boolean based on whether type is integer.
        """
        if "type" in attrentry:
            if isinstance(attrentry["type"], list):
                for item in attrentry["type"]:
                    if item.lower() == "integer" or item.lower() == "number":
                        return True
            elif attrentry["type"] == "array":
                for key, value in attrentry["items"].items():
                    if key.lower() == "type":
                        if value.lower() == "integer" or value.lower() == "number":
                            return True
            else:
                if attrentry["type"].lower() == "integer" or attrentry["type"].lower().lower() == "number":
                    return True
        elif "Type" in attrentry:
            if attrentry["Type"].lower() == "integer":
                return True

        return False

    def validate(self, newvallist, _):
        """Validate against schemas.

        :param newvallist: New value to be used for validation in a list
        :type newvallist: list
        :returns: An error if validation fails criteria.
        """
        result = list()
        try:
            intval = int(newvallist[0])
            newvallist[0] = intval
        except:
            result.append(RegistryValidationError("'%(Name)s' must " "be an integer value'" % (self), regentry=self))
            return result

        if newvallist[0] and not str(intval).isdigit():
            result.append(RegistryValidationError("'%(Name)s' must " "be an integer value'" % (self), regentry=self))
            return result

        if "LowerBound" in self:
            if intval < int(self["LowerBound"]):
                result.append(
                    RegistryValidationError(
                        "'%s' must be greater" " than or equal to '%s'" % (self.Name, int(self["LowerBound"])),
                        regentry=self,
                    )
                )

        if "UpperBound" in self:
            if intval > int(self["UpperBound"]):
                result.append(
                    RegistryValidationError(
                        "'%s' must be less " "than or equal to '%s'" % (self.Name, int(self["LowerBound"])),
                        regentry=self,
                    )
                )

        return result

    def print_help(self, name):
        """Human readable schema information specific to Integer data.

        :param name: Clean name for outputting human readable info.
        :type name: str
        :returns: A human readable string of schema data.
        """
        outdata = self.common_print_help(name)
        return outdata


class ObjectValidator(BaseValidator):
    """Object validator class"""

    def __init__(self, d):
        super(ObjectValidator, self).__init__(d)

    @staticmethod
    def is_type(attrentry):
        """Validate that the type is object.

        :param attrentry: Registry model entry used for validation.
        :type attrentry: dict
        :returns: A boolean based on whether type is object.
        """
        if "type" in attrentry:
            if isinstance(attrentry["type"], list):
                for item in attrentry["type"]:
                    if item.lower() == "object":
                        return True
            elif attrentry["type"] == "array":
                for key, value in attrentry["items"].items():
                    if key.lower() == "type" and value.lower() == "object":
                        return True
                    elif key.lower() == "anyof":
                        try:
                            if value[0]["type"] == "object":
                                return True
                        except Exception:
                            continue
            else:
                if attrentry["type"].lower() == "object":
                    return True
        elif "Type" in attrentry:
            if attrentry["Type"].lower() == "object":
                return True

        return False

    def validate(self, newval, name):
        """Validate against schemas.

        :param newval: New value to be used for validation in a list
        :type newval: list
        :param name: Clean name for outputting human readable info.
        :type name: str
        :returns: An error if value is invalid.
        """
        # TODO: need to add logic for true postive and false negatives.
        result = list()
        if isinstance(newval[0], (dict, six.string_types, int)):
            result.append(
                RegistryValidationError(
                    "'%s' is not a valid setting for '%s'" % (newval[0], name),
                    regentry=self,
                )
            )
        return result

    def print_help(self, name):
        """Human readable schema information specific to Object data.

        :param name: Clean name for outputting human readable info.
        :type name: str
        :returns: A human readable string of schema data.
        """
        wrapper = textwrap.TextWrapper()
        wrapper.initial_indent = " " * 4
        wrapper.subsequent_indent = " " * 4
        outdata = self.common_print_help(name)
        if "properties" in self:
            outdata += "\nSUB-PROPERTIES\n"
            propdata = ", ".join(self.properties.keys())
            outdata += "%s" % wrapper.fill("%s" % propdata)
            outdata += "\n"
        elif "items" in self:
            outdata += "\nSUB-PROPERTIES\n"
            propdata = ", ".join(self["items"].properties.keys())
            outdata += "%s" % wrapper.fill("%s" % propdata)
            outdata += "\n"

        return outdata


class PasswordValidator(BaseValidator):
    """Password validator class"""

    def __init__(self, d):
        super(PasswordValidator, self).__init__(d)

    @staticmethod
    def is_type(attrentry):
        """Validate that the type is a password.

        :param attrentry: Registry model entry used for validation.
        :type attrentry: dict
        :returns: A boolean based on whether type is a password.
        """
        if "type" in attrentry:
            if isinstance(attrentry["type"], list):
                for item in attrentry["type"]:
                    if item.lower() == "password":
                        return True
            elif attrentry["type"] == "array":
                for key, value in attrentry["items"].items():
                    if key.lower() == "type" and value.lower() == "password":
                        return True
            else:
                if attrentry["type"].lower() == "password":
                    return True
        elif "Type" in attrentry:
            if attrentry["Type"].lower() == "password":
                return True

        return False

    def validate(self, newvallist, _):
        """Validate against schemas.

        :param newvallist: New value to be used for validation in a list
        :type newvallist: list
        :returns: An error if validation fails criteria.
        """
        result = list()
        newval = newvallist[0]

        if newval is None:
            return result

        if not isinstance(newval, str):
            result.append(RegistryValidationError("Given value must be a string"))
        if "MinLength" in self:
            if len(newval) < int(self["MinLength"]):
                result.append(
                    RegistryValidationError(
                        "'%s' must be at least" " '%s' characters long" % (self.Name, int(self["MinLength"])),
                        regentry=self,
                    )
                )

        if "MaxLength" in self:
            if len(newval) > int(self["MaxLength"]):
                result.append(
                    RegistryValidationError(
                        "'%s' must be less " "than '%s' characters long" % (self.Name, int(self["MaxLength"])),
                        regentry=self,
                    )
                )

        if "ValueExpression" in self:
            if self["ValueExpression"]:
                pat = re.compile(self["ValueExpression"])
                if newval and not pat.match(newval):
                    result.append(
                        RegistryValidationError(
                            "'%(Name)s' must " "match the regular expression '%(Value" "Expression)s'" % (self),
                            regentry=self,
                        )
                    )

        return result

    def print_help(self, name):
        """Human readable schema information specific to Password data.

        :param name: Clean name for outputting human readable info.
        :type name: str
        :returns: A human readable string of schema data.
        """
        wrapper = textwrap.TextWrapper()
        wrapper.initial_indent = " " * 4
        wrapper.subsequent_indent = " " * 4
        outdata = self.common_print_help(name)
        if "MinLength" in self:
            outdata += "\nMIN LENGTH\n"
            outdata += "%s" % wrapper.fill("%(MinLength)s" % self)
            outdata += "\n"

        if "MaxLength" in self:
            outdata += "\nMAX LENGTH\n"
            outdata += "%s" % wrapper.fill("%(MaxLength)s" % self)
            outdata += "\n"
        return outdata


class Typepathforval(object):
    """Way to store the typepath defines object."""

    typepath = None

    def __new__(cls, typepathobj):
        if typepathobj:
            Typepathforval.typepath = typepathobj


def checkattr(aobj, prop):
    """Check if an attribute exists"""
    try:
        if hasattr(aobj, prop):
            return True
    except:
        pass
    return False



 # Copyright 2020 Hewlett Packard Enterprise Development LP
 #
 # Licensed under the Apache License, Version 2.0 (the "License"); you may
 # not use this file except in compliance with the License. You may obtain
 # a copy of the License at
 #
 #      http://www.apache.org/licenses/LICENSE-2.0
 #
 # Unless required by applicable law or agreed to in writing, software
 # distributed under the License is distributed on an "AS IS" BASIS, WITHOUT
 # WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. See the
 # License for the specific language governing permissions and limitations
 # under the License.

# -*- coding: utf-8 -*-
"""
An example of changing the temporary boot order
"""

import sys
import json
from redfish import RedfishClient
from redfish.rest.v1 import ServerDownOrUnreachableError

from get_resource_directory import get_resource_directory

def change_temporary_boot_order(_redfishobj, boottarget):

    systems_members_uri = None
    systems_members_response = None

    resource_instances = get_resource_directory(_redfishobj)
    if DISABLE_RESOURCE_DIR or not resource_instances:
        systems_uri = _redfishobj.root.obj['Systems']['@odata.id']
        systems_response = _redfishobj.get(systems_uri)
        systems_members_uri = next(iter(systems_response.obj['Members']))['@odata.id']
        systems_members_response = _redfishobj.get(systems_members_uri)
    else:
        for instance in resource_instances:
            if '#ComputerSystem.' in instance['@odata.type']:
                systems_members_uri = instance['@odata.id']
                systems_members_response = _redfishobj.get(systems_members_uri)

    if systems_members_response:
        print("\n\nShowing BIOS attributes before changes:\n\n")
        print(json.dumps(systems_members_response.dict.get('Boot'), indent=4, sort_keys=True))
    body = {'Boot': {'BootSourceOverrideTarget': boottarget}}
    resp = _redfishobj.patch(systems_members_uri, body)

    #If iLO responds with soemthing outside of 200 or 201 then lets check the iLO extended info
    #error message to see what went wrong
    if resp.status == 400:
        try:
            print(json.dumps(resp.obj['error']['@Message.ExtendedInfo'], indent=4, sort_keys=True))
        except Exception as excp:
            sys.stderr.write("A response error occurred, unable to access iLO Extended Message "\
                             "Info...")
    elif resp.status != 200:
        sys.stderr.write("An http response of \'%s\' was returned.\n" % resp.status)
    else:
        print("\nSuccess!\n")
        print(json.dumps(resp.dict, indent=4, sort_keys=True))
        if systems_members_response:
            print("\n\nShowing boot override target:\n\n")
            print(json.dumps(systems_members_response.dict.get('Boot'), indent=4, sort_keys=True))

if __name__ == "__main__":
    # When running on the server locally use the following commented values
    #SYSTEM_URL = None
    #LOGIN_ACCOUNT = None
    #LOGIN_PASSWORD = None

    # When running remotely connect using the secured (https://) address,
    # account name, and password to send https requests
    # SYSTEM_URL acceptable examples:
    # "https://10.0.0.100"
    # "https://ilo.hostname"
    SYSTEM_URL = "https://10.0.0.100"
    LOGIN_ACCOUNT = "admin"
    LOGIN_PASSWORD = "password"

    # string to represent the selected temporary boot device
    TEMP_DEVICE = "Hdd"
    # flag to force disable resource directory. Resource directory and associated operations are
    # intended for HPE servers.
    DISABLE_RESOURCE_DIR = False

    try:
        # Create a Redfish client object
        REDFISHOBJ = RedfishClient(base_url=SYSTEM_URL, username=LOGIN_ACCOUNT, \
                                                                            password=LOGIN_PASSWORD)
        # Login with the Redfish client
        REDFISHOBJ.login()
    except ServerDownOrUnreachableError as excp:
        sys.stderr.write("ERROR: server not reachable or does not support RedFish.\n")
        sys.exit()

    change_temporary_boot_order(REDFISHOBJ, TEMP_DEVICE)
    REDFISHOBJ.logout()



###
# Copyright 2020 Hewlett Packard Enterprise, Inc. All rights reserved.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#  http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
###

# -*- coding: utf-8 -*-
"""A quickstart example for RedfishClient"""

import sys
import redfish

# When running on the server locally use the following commented values
# HOST = "blobstore://."
# LOGIN_ACCOUNT = "None"
# LOGIN_PASSWORD = "None"

# When running remotely connect using the iLO address, iLO account name,
# and password to send https requests
SYSTEM_URL = "https://10.0.0.100"
LOGIN_ACCOUNT = "admin"
LOGIN_PASSWORD = "password"

## Create a REDFISH object
REDFISH_OBJ = redfish.RedfishClient(base_url=SYSTEM_URL, username=LOGIN_ACCOUNT,\
                                    password=LOGIN_PASSWORD)

# Login into the server and create a session
REDFISH_OBJ.login()

# Do a GET on a given path
RESPONSE = REDFISH_OBJ.get("/redfish/v1/systems/1")

# Print out the response
sys.stdout.write("%s\n" % RESPONSE)

# Logout of the current session
REDFISH_OBJ.logout()



 # Copyright 2020 Hewlett Packard Enterprise Development LP
 #
 # Licensed under the Apache License, Version 2.0 (the "License"); you may
 # not use this file except in compliance with the License. You may obtain
 # a copy of the License at
 #
 #      http://www.apache.org/licenses/LICENSE-2.0
 #
 # Unless required by applicable law or agreed to in writing, software
 # distributed under the License is distributed on an "AS IS" BASIS, WITHOUT
 # WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. See the
 # License for the specific language governing permissions and limitations
 # under the License.

# -*- coding: utf-8 -*-
"""
An example of configuring SNMP alert for HPE iLO systems
Usage: python <script>.py <ilo_addr> <ilo_user> <ilo_pass>
"""

import sys
import json
from redfish import RedfishClient
from redfish.rest.v1 import ServerDownOrUnreachableError

from ilorest_util import get_resource_directory
from ilorest_util import get_gen

def configure_snmp(_redfishobj, read_communities, snmp_alertdestinations, DISABLE_RESOURCE_DIR):
    snmp_service_uri = None
    resource_instances = get_resource_directory(_redfishobj)
    if DISABLE_RESOURCE_DIR or not resource_instances:
        #if we do not have a resource directory or want to force it's non use to find the
        #relevant URI
        managers_uri = _redfishobj.root.obj['Managers']['@odata.id']
        managers_response = _redfishobj.get(managers_uri)
        managers_members_uri = next(iter(managers_response.obj['Members']))['@odata.id']
        managers_members_response = _redfishobj.get(managers_members_uri)
        snmp_service_uri = managers_members_response.obj.Oem.Hpe.Links['Snmp']['@odata.id']
    else:
        for instance in resource_instances:
            #Use Resource directory to find the relevant URI
            if '#HpeiLOSnmpService.' in instance['@odata.type']:
                snmp_service_uri = instance['@odata.id']

    if snmp_service_uri:
        #body = {"AlertsEnabled": snmp_alerts, "ReadCommunities": read_communities}
        body = {"AlertDestinations": snmp_alertdestinations}
        resp = _redfishobj.patch(snmp_service_uri, body)

        #If iLO responds with soemthing outside of 200 or 201 then lets check the iLO extended info
        #error message to see what went wrong
        if resp.status == 400:
            try:
                print(json.dumps(resp.obj['error']['@Message.ExtendedInfo'], indent=4, \
                                                                                                                                                    sort_keys=True))
            except Exception as excp:
                sys.stderr.write("A response error occurred, unable to access iLO Extended "\
                                                     "Message Info...")
        elif resp.status != 200:
            sys.stderr.write("An http response of \'%s\' was returned.\n" % resp.status)
        else:
            print("Success!\n")
            print(json.dumps(resp.dict, indent=4, sort_keys=True))

def set_snmp_alert_destination(_redfishobj, snmp_service_uri, alert_destination_list):
    data = _redfishobj.get(snmp_service_uri)
    if data.dict.get("AlertDestinations"):
        resp = _redfishobj.patch(snmp_service_uri, {'AlertDestinations': alert_destination_list})
    else:
        raise Exception("\'AlertDestinations\' property is not available/modifyable.\n")
        #If iLO responds with soemthing outside of 200 or 201 then lets check the iLO extended info
        #error message to see what went wrong
    if resp.status == 400:
        try:
            print(json.dumps(resp.obj['error']['@Message.ExtendedInfo'], indent=4, sort_keys=True))
        except Exception as excp:
            sys.stderr.write("A response error occurred, unable to access iLO Extended " "Message Info...")
    elif resp.status != 200:
        sys.stderr.write("An http response of \'%s\' was returned.\n" % resp.status)
    else:
        print("\nPatch operation successful!\n\nResponse:")
        print(json.dumps(resp.dict, indent=4, sort_keys=True))
        snmp_service_response = _redfishobj.get(snmp_service_uri).dict.get('AlertDestinations')
        print("\n\nPrinting updated SNMP alert destination:\n")
        print(json.dumps(snmp_service_response, indent=4, sort_keys=True))


if __name__ == "__main__":
    # When running on the server locally use the following commented values
    #SYSTEM_URL = None
    #LOGIN_ACCOUNT = None
    #LOGIN_PASSWORD = None

    # When running remotely connect using the secured (https://) address,
    # account name, and password to send https requests
    # SYSTEM_URL acceptable examples:
    # "https://10.0.0.100"
    # "https://ilo.hostname"
    #SYSTEM_URL = "https://15.146.46.45"
    #LOGIN_ACCOUNT = "admin"
    #LOGIN_PASSWORD = "admin123"

    SYSTEM_URL = sys.argv[1]
    LOGIN_ACCOUNT = sys.argv[2]
    LOGIN_PASSWORD = sys.argv[3]

    #Properties:
    #read communities array
    READ_COMMUNITIES = ["public", "", ""]
    #alerts_enabled primitive (boolean)
    #ALERTS_ENABLED = True
    #Alert Destination
    ALERTS_DESTINATION = ["1.1.1.1","2.2.2.2"]
    # flag to force disable resource directory. Resource directory and associated operations are
    # intended for HPE servers.
    DISABLE_RESOURCE_DIR = False
    snmp_service_uri = "/redfish/v1/Managers/1/SNMPService/"

    # Number of max alert destination supported on iLO4 and iLO5 is different
    alert_destination_list = ["ILOCN771702NJ", "15.146.46.55" , "15.146.46.58"]

    try:
        # Create a Redfish client object
        REDFISHOBJ = RedfishClient(base_url=SYSTEM_URL, username=LOGIN_ACCOUNT, \
                                                                            password=LOGIN_PASSWORD)
        # Login with the Redfish client
        REDFISHOBJ.login()
    except ServerDownOrUnreachableError as excp:
        sys.stderr.write("ERROR: server not reachable or does not support RedFish.\n")
        sys.exit()
    (ilogen,_) = get_gen(REDFISHOBJ)
    print ("Generation is ", ilogen)
    if int(ilogen) == 5:
        configure_snmp(REDFISHOBJ, READ_COMMUNITIES, ALERTS_DESTINATION, DISABLE_RESOURCE_DIR)
    else:
        set_snmp_alert_destination(REDFISHOBJ, snmp_service_uri, alert_destination_list)
    REDFISHOBJ.logout()
